# -*- coding: utf-8 -*-
# Author: Mehedi Hasan Khondoker
# Email: mehedihasankhondoker [at] gmail.com
# Copyright (c) 2024, Mehedi Hasan Khondoker.  All Rights Reserved.

# This library is build for Bangladesh format Number to Word conversion.
# You are welcome as contributor to the library.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.


from decimal import Decimal

RANKING = ['', 'প্রথম', 'দ্বিতীয়', 'তৃতীয়', 'চতুর্থ', 'পঞ্চম', 'ষষ্ঠ',
           'সপ্তম', 'অষ্টম', 'নবম', 'দশম']  # pragma: no cover

AKOK = ['', 'এক', 'দুই', 'তিন', 'চার', 'পাঁচ', 'ছয়',
        'সাত', 'আট', 'নয়']  # pragma: no cover

DOSOK = [
    'দশ', 'এগারো', 'বারো', 'তেরো', 'চৌদ্দ', 'পনের',
    'ষোল', 'সতের', 'আঠারো', 'উনিশ',
    'বিশ', 'একুশ', 'বাইশ', 'তেইশ', 'চব্বিশ', 'পঁচিশ',
    'ছাব্বিশ', 'সাতাশ', 'আটাশ', 'উনত্রিশ',
    'ত্রিশ', 'একত্রিশ', 'বত্রিশ', 'তেত্রিশ', 'চৌত্রিশ', 'পঁইত্রিশ',
    'ছত্রিশ', 'সাতত্রিশ', 'আটত্রিশ', 'উনচল্লিশ', 'চল্লিশ',
    'একচল্লিশ', 'বিয়াল্লিশ', 'তেতাল্লিশ', 'চৌচল্লিশ',
    'পঁয়তাল্লিশ', 'ছেচল্লিশ', 'সাতচল্লিশ', 'আটচল্লিশ', 'উনপঞ্চাশ',
    'পঞ্চাশ', 'একান্ন', 'বাহান্ন', 'তিপ্পান্ন', 'চুয়ান্ন', 'পঞ্চান্ন',
    'ছাপ্পান্ন', 'সাতান্ন', 'আটান্ন', 'উনষাট', 'ষাট',
    'একষট্টি', 'বাষট্টি', 'তেষট্টি', 'চৌষট্টি', 'পঁয়ষট্টি',
    'ছিষট্টি', 'সাতষট্টি', 'আটষট্টি', 'উনসত্তর', 'সত্তর',
    'একাত্তর ', 'বাহাত্তর', 'তিয়াত্তর', 'চুয়াত্তর', 'পঁচাত্তর',
    'ছিয়াত্তর', 'সাতাত্তর', 'আটাত্তর', 'উনআশি', 'আশি',
    'একাশি', 'বিরাশি', 'তিরাশি', 'চুরাশি', 'পঁচাশি',
    'ছিয়াশি', 'সাতাশি', 'আটাশি', 'উননব্বই', 'নব্বই',
    'একানব্বই', 'বিরানব্বই', 'তিরানব্বই', 'চুরানব্বই', 'পঁচানব্বই',
    'ছিয়ানব্বই', 'সাতানব্বই', 'আটানব্বই', 'নিরানব্বই'
]  # pragma: no cover
HAZAR = ' হাজার '  # pragma: no cover
LAKH = ' লাখ '  # pragma: no cover
KOTI = ' কোটি '  # pragma: no cover
MAX_NUMBER = 9999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999  # noqa: E501 # pragma: no cover


class NumberTooLargeError(Exception):
    """Custom exception for numbers that are too large."""
    pass


class Num2Word_BN:

    @staticmethod
    def str_to_number(number):
        return abs(Decimal(str(number)))  # pragma: no cover

    @staticmethod
    def parse_number(number: Decimal):
        dosomik = str(number - int(number)).split('.')[1:]
        dosomik_str = ''.join(dosomik) if dosomik else 0
        return int(number), int(dosomik_str)

    @staticmethod
    def parse_paisa(number: Decimal):
        # 1-99 for paisa count so two digits are valid.
        paisa = str(number - int(number)).split('.')[1:]
        paisa_str = ''.join(paisa) if paisa else 0

        # this is need, when we parse to decimal it removes trailing 0 .
        if paisa_str:
            paisa_str = str(int(paisa_str) * 100)[:2]
        return int(number), int(paisa_str)

    def _is_smaller_than_max_number(self, number):
        if MAX_NUMBER >= number:
            return True
        raise NumberTooLargeError(f'Too Large number maximum '
                                  f'value={MAX_NUMBER}')

    def _dosomik_to_bengali_word(self, number):
        word = ''
        for i in str(number):
            word += ' ' + AKOK[int(i)]
        return word

    def _number_to_bengali_word(self, number):
        if number == 0:
            return 'শূন্য'

        words = ''

        if number >= 10 ** 7:
            words += self._number_to_bengali_word(number // 10 ** 7) + KOTI
            number %= 10 ** 7

        if number >= 10 ** 5:
            words += self._number_to_bengali_word(number // 10 ** 5) + LAKH
            number %= 10 ** 5

        if number >= 1000:
            words += self._number_to_bengali_word(number // 1000) + HAZAR
            number %= 1000

        if number >= 100:
            words += AKOK[number // 100] + 'শত '
            number %= 100

        if 10 <= number <= 99:
            words += DOSOK[number - 10] + ' '
            number = 0

        if 0 < number < 10:
            words += AKOK[number] + ' '

        return words.strip()

    def to_currency(self, val):
        """
        This function represent a number to word in bangla taka and paisa.
        example:
        1 = এক টাকা,
        101 = একশত এক টাকা,
        9999.15 = নয় হাজার নয়শত নিরানব্বই টাকা পনের পয়সা
        and so on.
        """

        dosomik_word = None
        number = self.str_to_number(val)
        number, decimal_part = self.parse_paisa(number)
        self._is_smaller_than_max_number(number)

        if decimal_part > 0:
            dosomik_word = f' {self._number_to_bengali_word(decimal_part)} পয়সা'  # noqa: E501

        words = f'{self._number_to_bengali_word(number)} টাকা'

        if dosomik_word:
            return (words + dosomik_word).strip()
        return words.strip()

    def to_cardinal(self, number):
        """
        This function represent a number to word in bangla.
        example:
        1 = এক,
        101 = একশত এক,
        9999 = নয় হাজার নয়শত নিরানব্বই
        and so on.
        """

        dosomik_word = None
        number = self.str_to_number(number)
        number, decimal_part = self.parse_number(number)
        self._is_smaller_than_max_number(number)

        if decimal_part > 0:
            dosomik_word = f' দশমিক{self._dosomik_to_bengali_word(decimal_part)}'  # noqa: E501

        words = self._number_to_bengali_word(number)

        if dosomik_word:
            return (words + dosomik_word).strip()
        return words.strip()

    def to_ordinal(self, number):
        return self.to_cardinal(number)

    def to_ordinal_num(self, number):
        """
        This function represent a number to ranking in bangla.
        example:
        1 = প্রথম,
        2 = দ্বিতীয়,
        1001 = এক হাজার একতম
        and so on.
        """
        self._is_smaller_than_max_number(number)

        if number in range(1, 11):
            return RANKING[number]
        else:
            rank = self.to_cardinal(int(abs(number)))
            if rank.endswith('ত'):
                return rank + 'ম'
            return rank + 'তম'

    def to_year(self, number):
        """
        This function represent a number to year in bangla.
        example:
        2002 = দুই হাজার দুই সাল,
        2024 = দুই হাজার চব্বিশ সাল
        and so on.
        """
        self._is_smaller_than_max_number(number)
        return self.to_cardinal(int(abs(number))) + ' সাল'



# -*- encoding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .lang_EU import Num2Word_EU


class Num2Word_KN(Num2Word_EU):
    def set_high_numwords(self, high):
        for n, word in self.high_numwords:
            self.cards[10**n] = word

    def setup(self):
        self.low_numwords = [
            "ತೊಂಬತ್ತೊಂಬತ್ತು",
            "ತೊಂಬತ್ತೆಂಟು",
            "ತೊಂಬತ್ತೇಳು",
            "ತೊಂಬತ್ತಾರು",
            "ತೊಂಬತ್ತೈದು",
            "ತೊಂಬತ್ತ ನಾಲ್ಕು",
            "ತೊಂಬತ್ತ ಮೂರು",
            "ತೊಂಬತ್ತೆರಡು",
            "ತೊಂಬತ್ತೊಂದು",
            "ತೊಂಬತ್ತು",
            "ಎಂಬತ್ತೊಂಬತ್ತು",
            "ಎಂಬತ್ತೆಂಟು",
            "ಎಂಬತ್ತೇಳು",
            "ಎಂಬತ್ತಾರು",
            "ಎಂಬತ್ತೈದು",
            "ಎಂಬತ್ತ್ ನಾಲ್ಕು",
            "ಎಂಬತ್ತ್ ಮೂರು",
            "ಎಂಬತ್ತೆರಡು",
            "ಎಂಬತ್ತೊಂದು",
            "ಎಂಬತ್ತು",
            "ಎಪ್ಪತ್ತೊಂಬತ್ತು",
            "ಎಪ್ಪತ್ತೆಂಟು",
            "ಎಪ್ಪತ್ತೇಳು",
            "ಎಪ್ಪತ್ತಾರು",
            "ಎಪ್ಪತ್ತೈದು",
            "ಎಪ್ಪತ್ತ್ ನಾಲ್ಕು",
            "ಎಪ್ಪತ್ತ್ ಮೂರು",
            "ಎಪ್ಪತ್ತೆರಡು",
            "ಎಪ್ಪತ್ತೊಂದು",
            "ಎಪ್ಪತ್ತು",
            "ಅರವತ್ತೊಂಬತ್ತು",
            "ಅರವತ್ತೆಂಟು",
            "ಅರವತ್ತೇಳು",
            "ಅರವತ್ತಾರು",
            "ಅರವತ್ತೈದು",
            "ಅರವತ್ತ್ ನಾಲ್ಕು",
            "ಅರವತ್ತ್ ಮೂರು",
            "ಅರವತ್ತೆರಡು",
            "ಅರವತ್ತೊಂದು",
            "ಅರವತ್ತು",
            "ಐವತ್ತೊಂಬತ್ತು",
            "ಐವತ್ತೆಂಟು",
            "ಐವತ್ತೇಳು",
            "ಐವತ್ತಾರು",
            "ಐವತ್ತೈದು",
            "ಐವತ್ತ್ನಾಲ್ಕು",
            "ಐವತ್ತಮೂರು",
            "ಐವತ್ತೆರಡು",
            "ಐವತ್ತೊಂದು",
            "ಐವತ್ತು",
            "ನಲವತ್ತೊಂಬತ್ತು",
            "ನಲವತ್ತೆಂಟು",
            "ನಲವತ್ತೇಳು",
            "ನಲವತ್ತಾರು",
            "ನಲವತ್ತೈದು",
            "ನಲವತ್ತ್ ನಾಲ್ಕು",
            "ನಲವತ್ತ್ ಮೂರು",
            "ನಲವತ್ತ್ ಎರಡು",
            "ನಲವತ್ತೊಂದು",
            "ನಲವತ್ತು",
            "ಮೂವತ್ತ್ ಒಂಬತ್ತು",
            "ಮೂವತ್ಎಂಟು",
            "ಮೂವತ್ಏಳು",
            "ಮೂವತ್ಆರು",
            "ಮೂವತ್ತ್ ಐದು",
            "ಮೂವತ್ತ್ ನಾಲ್ಕು",
            "ಮೂವತ್ತ್ ಮೂರು",
            "ಮೂವತ್ತ್ಎರಡು",
            "ಮೂವತ್ತ್ಒಂದು",
            "ಮೂವತ್ತು",
            "ಇಪ್ಪತ್ತ್ಒಂಬತ್ತು",
            "ಇಪ್ಪತ್ತ್ಎಂಟು",
            "ಇಪ್ಪತ್ತ್ಏಳು",
            "ಇಪ್ಪತ್ತ್ಆರು",
            "ಇಪ್ಪತ್ತ್ ಐದು",
            "ಇಪ್ಪತ್ತ್ ನಾಲ್ಕು",
            "ಇಪ್ಪತ್ತ್ ಮೂರು",
            "ಇಪ್ಪತ್ತ್ ಎರಡು",
            "ಇಪ್ಪತ್ತ್ ಒಂದು",
            "ಇಪ್ಪತ್ತು",
            "ಹತ್ತೊಂಬತ್ತು",
            "ಹದಿನೆಂಟು",
            "ಹದಿನೇಳು",
            "ಹದಿನಾರು",
            "ಹದಿನೈದು",
            "ಹದಿನಾಲ್ಕು",
            "ಹದಿಮೂರು",
            "ಹನ್ನೆರಡು",
            "ಹನ್ನೊಂದು",
            "ಹತ್ತು",
            "ಒಂಬತ್ತು",
            "ಎಂಟು",
            "ಏಳು",
            "ಆರು",
            "ಐದು",
            "ನಾಲ್ಕು",
            "ಮೂರು",
            "ಎರಡು",
            "ಒಂದು",
            "ಸೊನ್ನೆ",
        ]

        self.mid_numwords = [(100, "ನೂರು")]

        self.high_numwords = [(7, "ಕೋಟಿ"), (5, "ಒಂದು ಲಕ್ಷ"), (3, "ಸಾವಿರ")]

        self.pointword = "ಬಿಂದು"

        self.modifiers = [
            "್",
            "ಾ",
            "ಿ",
            "ೀ",
            "ೀ",
            "ು",
            "ೂ",
            "ೃ",
            "ೆ",
            "ೇ",
            "ೈ",
            "ೊ",
            "ೋ",
            "ೌ",
            "ೕ",
        ]

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif 100 > lnum > rnum:
            return ("%s-%s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 100 > rnum:
            if ltext[-1] in self.modifiers:
                return ("%s %s" % (ltext[:-1], rtext), lnum + rnum)
            else:
                return ("%s %s" % (ltext + "ದ", rtext), lnum + rnum)
        elif rnum > lnum:
            return ("%s %s" % (ltext, rtext), lnum * rnum)
        return ("%s %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s%s" % (value, self.to_ordinal(value))

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outwords = self.to_cardinal(value)
        if outwords[-1] in self.modifiers:
            outwords = outwords[:-1]
        ordinal_num = outwords + "ನೇ"
        return ordinal_num



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA


def splitbyx(n, x, format_int=True):
    length = len(n)
    if length > x:
        start = length % x
        if start > 0:
            result = n[:start]
            yield int(result) if format_int else result
        for i in range(start, length, x):
            result = n[i:i+x]
            yield int(result) if format_int else result
    else:
        yield int(n) if format_int else n


def get_digits(n):
    a = [int(x) for x in reversed(list(('%03d' % n)[-3:]))]
    return a



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import print_function, unicode_literals

from .lang_EU import Num2Word_EU


class Num2Word_NL(Num2Word_EU):
    CURRENCY_FORMS = {
        'EUR': (('euro', 'euro'), ('cent', 'cent')),
        'GBP': (('pond', 'pond'), ('penny', 'pence')),
        'USD': (('dollar', 'dollar'), ('cent', 'cent')),
        'CNY': (('yuan', 'yuan'), ('jiao', 'fen')),
    }

    GIGA_SUFFIX = "iljard"
    MEGA_SUFFIX = "iljoen"

    def setup(self):
        super(Num2Word_NL, self).setup()

        self.negword = "min "
        self.pointword = "komma"
        # "Cannot treat float %s as ordinal."
        self.errmsg_floatord = (
            "Het zwevende puntnummer %s kan niet omgezet worden " +
            "naar een ordernummer."
        )
        # "type(((type(%s)) ) not in [long, int, float]"
        self.errmsg_nonnum = (
            "Alleen nummers (type (%s)) kunnen naar " +
            "woorden omgezet worden."
        )
        # "Cannot treat negative num %s as ordinal."
        self.errmsg_negord = (
            "Het negatieve getal %s kan niet omgezet " +
            "worden naar een ordernummer."
        )
        # "abs(%s) must be less than %s."
        self.errmsg_toobig = "Het getal %s moet minder zijn dan %s."
        self.exclude_title = []

        lows = ["non", "okt", "sept", "sext", "quint", "quadr", "tr", "b", "m"]
        units = ["", "un", "duo", "tre", "quattuor", "quin", "sex", "sept",
                 "okto", "novem"]
        tens = ["dez", "vigint", "trigint", "quadragint", "quinquagint",
                "sexagint", "septuagint", "oktogint", "nonagint"]

        self.high_numwords = (
            ["zend"] + self.gen_high_numwords(units, tens, lows))

        self.mid_numwords = [(1000, "duizend"), (100, "honderd"),
                             (90, "negentig"), (80, "tachtig"),
                             (70, "zeventig"), (60, "zestig"),
                             (50, "vijftig"), (40, "veertig"),
                             (30, "dertig")]

        self.low_numwords = ["twintig", "negentien", "achttien", "zeventien",
                             "zestien", "vijftien", "veertien", "dertien",
                             "twaalf", "elf", "tien", "negen", "acht", "zeven",
                             "zes", "vijf", "vier", "drie", "twee", "één",
                             "nul"]

        # Wiktionary says it is "nulde", not "nulte" or "nule"
        # https://en.wiktionary.org/wiki/nulde
        self.ords = {"nul": "nuld",
                     "één": "eerst",
                     "twee": "tweed",
                     "drie": "derd",
                     "vier": "vierd",
                     "vijf": "vijfd",
                     "zes": "zesd",
                     "zeven": "zevend",
                     "acht": "achtst",
                     "negen": "negend",
                     "tien": "tiend",
                     "elf": "elfd",
                     "twaalf": "twaalfd",

                     "ig": "igst",
                     "erd": "erdst",
                     "end": "endst",
                     "joen": "joenst",
                     "rd": "rdst"}

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next

        if cnum == 1:
            if nnum < 10 ** 6:
                return next
            ctext = "een"

        if nnum > cnum:
            if nnum >= 10 ** 6:
                ctext += " "
            val = cnum * nnum
        else:
            if nnum < 10 < cnum < 100:
                if nnum == 1:
                    ntext = "een"

                if ntext.endswith("e"):
                    ntext += "ën"  # "n"
                else:
                    ntext += "en"
                ntext, ctext = ctext, ntext  # + "en"
            elif cnum >= 10 ** 6:
                ctext += " "
            val = cnum + nnum

        word = ctext + ntext
        return word, val

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outword = self.to_cardinal(value)
        for key in self.ords:
            if outword.endswith(key):
                outword = outword[:len(outword) - len(key)] + self.ords[key]
                break
        return outword + "e"

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return str(value) + "e"

    def pluralize(self, n, forms):
        """
        :param n:
        :param forms:
        :return:

        gettext form is nplurals=2; plural=(n != 1);
        but this claims https://onzetaal.nl/taaladvies/euro-euros/
        not sure if it's applied only to euro
        """
        return forms[0]

    def to_year(self, val, longval=True):
        if not (val // 100) % 10:
            return self.to_cardinal(val)
        return self.to_splitnum(val, hightxt="honderd", longval=longval)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from . import lang_EN


class Num2Word_EN_NG(lang_EN.Num2Word_EN):

    CURRENCY_FORMS = {'NGN': (('naira', 'naira'), ('kobo', 'kobo'))}

    CURRENCY_ADJECTIVES = {'NGN': 'Nigerian'}

    def to_currency(
        self, val, currency='NGN',
        kobo=True, separator=',',
        adjective=False
    ):
        result = super(Num2Word_EN_NG, self).to_currency(
            val, currency=currency, cents=kobo, separator=separator,
            adjective=adjective)
        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU


class Num2Word_NO(lang_EU.Num2Word_EU):
    GIGA_SUFFIX = "illard"
    MEGA_SUFFIX = "illion"
    CURRENCY_FORMS = {'NOK': (('krone', 'kroner'), ('øre', 'øre'))}

    def set_high_numwords(self, high):
        cap = 3 + 6 * len(high)

        for word, n in zip(high, range(cap, 3, -6)):
            if self.GIGA_SUFFIX:
                self.cards[10 ** n] = word + self.GIGA_SUFFIX

            if self.MEGA_SUFFIX:
                self.cards[10 ** (n - 3)] = word + self.MEGA_SUFFIX

    def setup(self):
        super(Num2Word_NO, self).setup()

        self.negword = "minus "
        self.pointword = "komma"
        self.exclude_title = ["og", "komma", "minus"]

        self.mid_numwords = [(1000, "tusen"), (100, "hundre"),
                             (90, "nitti"), (80, "\xe5tti"), (70, "sytti"),
                             (60, "seksti"), (50, "femti"), (40, "f\xf8rti"),
                             (30, "tretti")]
        self.low_numwords = ["tjue", "nitten", "atten", "sytten",
                             "seksten", "femten", "fjorten", "tretten",
                             "tolv", "elleve", "ti", "ni", "\xe5tte",
                             "syv", "seks", "fem", "fire", "tre", "to",
                             "en", "null"]
        self.ords_pl = {"to": "andre",
                        "tre": "tredje",
                        "fire": "fjerde",
                        "fem": "femte",
                        "seks": "sjette",
                        "syv": "syvende",
                        "\xe5tte": "\xe5ttende",
                        "ni": "niende",
                        "ti": "tiende",
                        "elleve": "ellevte",
                        "tolv": "tolvte",
                        "fjorten": "fjortende",
                        "femten": "femtende",
                        "seksten": "sekstende",
                        "sytten": "syttende",
                        "atten": "attende",
                        "nitten": "nittende",
                        "tjue": "tjuende",
                        "hundre": "hundrede",
                        "tusen": "tusende",
                        "million": "millionte"}
        # this needs to be done separately to not block 13-19 to_ordinal
        self.ords_sg = {"en": "f\xf8rste"}

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif 100 > lnum > rnum:
            return ("%s%s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 100 > rnum:
            return ("%s og %s" % (ltext, rtext), lnum + rnum)
        elif rnum > lnum:
            return ("%s %s" % (ltext, rtext), lnum * rnum)
        return ("%s %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outword = self.to_cardinal(value)
        for key in self.ords_pl:
            if outword.endswith(key):
                outword = outword[:len(outword) - len(key)] + self.ords_pl[key]
                break
        for key in self.ords_sg:
            if outword.endswith(key):
                outword = outword[:len(outword) - len(key)] + self.ords_sg[key]
                break
        return outword

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return str(value) + "."

    def to_year(self, val, longval=True):
        if not (val // 100) % 10:
            return self.to_cardinal(val)
        return self.to_splitnum(val, hightxt="hundre", jointxt="og",
                                longval=longval)

    def to_currency(self, val, currency='NOK', cents=True, separator=' og',
                    adjective=False):
        result = super(Num2Word_NO, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)

        # do not print "og null øre"
        result = result.replace(' og null øre', '')
        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2019, Filippo Costa.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .lang_EU import Num2Word_EU

# Globals
# -------

ZERO = "zero"

CARDINAL_WORDS = [
    ZERO, "uno", "due", "tre", "quattro", "cinque", "sei", "sette", "otto",
    "nove", "dieci", "undici", "dodici", "tredici", "quattordici", "quindici",
    "sedici", "diciassette", "diciotto", "diciannove"
]

ORDINAL_WORDS = [
    ZERO, "primo", "secondo", "terzo", "quarto", "quinto", "sesto", "settimo",
    "ottavo", "nono", "decimo", "undicesimo", "dodicesimo", "tredicesimo",
    "quattordicesimo", "quindicesimo", "sedicesimo", "diciassettesimo",
    "diciottesimo", "diciannovesimo"
]

# The script can extrapolate the missing numbers from the base forms.
STR_TENS = {2: "venti", 3: "trenta", 4: "quaranta", 6: "sessanta"}

# These prefixes are used for extremely big numbers.
EXPONENT_PREFIXES = [
    ZERO, "m", "b", "tr", "quadr", "quint", "sest", "sett", "ott", "nov", "dec"
]


GENERIC_DOLLARS = ('dollaro', 'dollari')
GENERIC_CENTS = ('centesimo', 'centesimi')
CURRENCIES_UNA = ('GBP')


# Main class
# ==========

class Num2Word_IT(Num2Word_EU):
    CURRENCY_FORMS = {
        'EUR': (('euro', 'euro'), GENERIC_CENTS),
        'USD': (GENERIC_DOLLARS, GENERIC_CENTS),
        'GBP': (('sterlina', 'sterline'), ('penny', 'penny')),
        'CNY': (('yuan', 'yuan'), ('fen', 'fen')),
    }
    MINUS_PREFIX_WORD = "meno "
    FLOAT_INFIX_WORD = " virgola "

    def setup(self):
        Num2Word_EU.setup(self)

    def __init__(self):
        pass

    def float_to_words(self, float_number, ordinal=False):
        if ordinal:
            prefix = self.to_ordinal(int(float_number))
        else:
            prefix = self.to_cardinal(int(float_number))
        float_part = str(float_number).split('.')[1]
        postfix = " ".join(
            # Drops the trailing zero and comma
            [self.to_cardinal(int(c)) for c in float_part]
        )
        return prefix + Num2Word_IT.FLOAT_INFIX_WORD + postfix

    def tens_to_cardinal(self, number):
        tens = number // 10
        units = number % 10
        if tens in STR_TENS:
            prefix = STR_TENS[tens]
        else:
            prefix = CARDINAL_WORDS[tens][:-1] + "anta"
        postfix = omitt_if_zero(CARDINAL_WORDS[units])
        return phonetic_contraction(prefix + postfix)

    def hundreds_to_cardinal(self, number):
        hundreds = number // 100
        prefix = "cento"
        if hundreds != 1:
            prefix = CARDINAL_WORDS[hundreds] + prefix
        postfix = omitt_if_zero(self.to_cardinal(number % 100))
        return phonetic_contraction(prefix + postfix)

    def thousands_to_cardinal(self, number):
        thousands = number // 1000
        if thousands == 1:
            prefix = "mille"
        else:
            prefix = self.to_cardinal(thousands) + "mila"
        postfix = omitt_if_zero(self.to_cardinal(number % 1000))
        # "mille" and "mila" don't need any phonetic contractions
        return prefix + postfix

    def big_number_to_cardinal(self, number):
        digits = [c for c in str(number)]
        length = len(digits)
        if length >= 66:
            raise NotImplementedError("The given number is too large.")
        # This is how many digits come before the "illion" term.
        #   cento miliardi => 3
        #   dieci milioni => 2
        #   un miliardo => 1
        predigits = length % 3 or 3
        multiplier = digits[:predigits]
        exponent = digits[predigits:]
        # Default infix string: "milione", "biliardo", "sestilione", ecc.
        infix = exponent_length_to_string(len(exponent))
        if multiplier == ["1"]:
            prefix = "un "
        else:
            prefix = self.to_cardinal(int("".join(multiplier)))
            # Plural form      ~~~~~~~~~~~
            infix = " " + infix[:-1] + "i"
        # Read as: Does the value of exponent equal 0?
        if set(exponent) != set("0"):
            postfix = self.to_cardinal(int("".join(exponent)))
            if " e " in postfix:
                infix += ", "
            else:
                infix += " e "
        else:
            postfix = ""
        return prefix + infix + postfix

    def to_cardinal(self, number):
        if number < 0:
            string = Num2Word_IT.MINUS_PREFIX_WORD + self.to_cardinal(-number)
        elif isinstance(number, float):
            string = self.float_to_words(number)
        elif number < 20:
            string = CARDINAL_WORDS[int(number)]
        elif number < 100:
            string = self.tens_to_cardinal(int(number))
        elif number < 1000:
            string = self.hundreds_to_cardinal(int(number))
        elif number < 1000000:
            string = self.thousands_to_cardinal(int(number))
        else:
            string = self.big_number_to_cardinal(number)
        return accentuate(string)

    def to_ordinal(self, number):
        tens = number % 100
        # Italian grammar is poorly defined here ¯\_(ツ)_/¯:
        #   centodecimo VS centodieciesimo VS centesimo decimo?
        is_outside_teens = not 10 < tens < 20
        if number < 0:
            return Num2Word_IT.MINUS_PREFIX_WORD + self.to_ordinal(-number)
        elif number % 1 != 0:
            return self.float_to_words(number, ordinal=True)
        elif number < 20:
            return ORDINAL_WORDS[int(number)]
        elif is_outside_teens and tens % 10 == 3:
            # Gets rid of the accent
            return self.to_cardinal(number)[:-1] + "eesimo"
        elif is_outside_teens and tens % 10 == 6:
            return self.to_cardinal(number) + "esimo"
        else:
            string = self.to_cardinal(number)[:-1]
            if string[-3:] == "mil":
                string += "l"
            return string + "esimo"

    def to_currency(self, val, currency='EUR', cents=True, separator=' e',
                    adjective=False):
        result = super(Num2Word_IT, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)
        # Handle exception. In italian language is "un euro",
        # "un dollaro" etc. (not "uno euro", "uno dollaro").
        # There is an exception, some currencies need "una":
        # e.g. "una sterlina"
        if currency in CURRENCIES_UNA:
            list_result = result.split(" ")
            if list_result[0] == "uno":
                list_result[0] = list_result[0].replace("uno", "una")
                result = " ".join(list_result)
        result = result.replace("uno", "un")
        return result

# Utils
# =====


def phonetic_contraction(string):
    return (string
            .replace("oo", "o")  # ex. "centootto"
            .replace("ao", "o")  # ex. "settantaotto"
            .replace("io", "o")  # ex. "ventiotto"
            .replace("au", "u")  # ex. "trentauno"
            .replace("iu", "u")  # ex. "ventiunesimo"
            )


def exponent_length_to_string(exponent_length):
    # We always assume `exponent` to be a multiple of 3. If it's not true, then
    # Num2Word_IT.big_number_to_cardinal did something wrong.
    prefix = EXPONENT_PREFIXES[exponent_length // 6]
    if exponent_length % 6 == 0:
        return prefix + "ilione"
    else:
        return prefix + "iliardo"


def accentuate(string):
    # This is inefficient: it may do several rewritings when deleting
    # half-sentence accents. However, it is the easiest method and speed is
    # not crucial (duh), so...
    return " ".join(
        # Deletes half-sentence accents and accentuates the last "tre"
        [w.replace("tré", "tre")[:-3] + "tré"
         # We shouldn't accentuate a single "tre": is has to be a composite
         # word.                ~~~~~~~~~~
         if w[-3:] == "tre" and len(w) > 3
         # Deletes half-sentence accents anyway
         #     ~~~~~~~~~~~~~~~~~~~~~~
         else w.replace("tré", "tre")
         for w in string.split()
         ])


def omitt_if_zero(number_to_string):
    return "" if number_to_string == ZERO else number_to_string



# -*- coding: utf-8 -*-Num2Word_TET
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA


from num2words.currency import parse_currency_parts

from .lang_EU import Num2Word_EU

DOLLAR = ('dolar', 'dolar')
CENTS = ('sentavu', 'sentavu')


class Num2Word_TET(Num2Word_EU):

    CURRENCY_FORMS = {
        'AUD': (DOLLAR, CENTS),
        'CAD': (DOLLAR, CENTS),
        'EUR': (('euro', 'euros'), CENTS),
        'GBP': (('pound sterling', 'pound sterling'), ('pence', 'pence')),
        'USD': (DOLLAR, CENTS),
    }

    GIGA_SUFFIX = None
    MEGA_SUFFIX = "iliaun"

    def setup(self):
        super().setup()
        lows = ["kuatr", "tr", "b", "m"]
        self.high_numwords = self.gen_high_numwords([], [], lows)
        self.negword = "menus "
        self.pointword = "vírgula"
        self.exclude_title = ["resin", "vírgula", "menus"]
        self.count = 0

        self.mid_numwords = [
            (1000, "rihun"), (100, "atus"), (90, "sianulu"),
            (80, "ualunulu"), (70, "hitunulu"), (60, "neenulu"),
            (50, "limanulu"), (40, "haatnulu"), (30, "tolunulu"),
            (20, "ruanulu")
        ]
        self.low_numwords = [
            "sanulu",
            "sia", "ualu", "hitu", "neen", "lima", "haat", "tolu", "rua",
            "ida", "mamuk"
        ]
        self.hundreds = {
            1: "atus",
            2: "atus rua",
            3: "atus tolu",
            4: "atus haat",
            5: "atus lima",
            6: "atus neen",
            7: "atus hitu",
            8: "atus ualu",
            9: "atus sia",
        }

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next

        if cnum == 1 and nnum < 100:
            return next

        if nnum < cnum:
            if nnum < 10:
                value_str = str(cnum + nnum)
                if int(value_str) > 100:
                    zero_list = value_str[1:-1]
                    all_zero = all(element == '0' for element in zero_list)
                    if all_zero:
                        if self.count >= 1:
                            self.count += 0
                            return (
                                "ho %s %s" % (ctext, ntext),
                                cnum + nnum
                            )
                        self.count += 1
                        return ("%s %s" % (ctext, ntext), cnum + nnum)

                return ("%s resin %s" % (ctext, ntext), cnum + nnum)
            else:
                return ("%s %s" % (ctext, ntext), cnum + nnum)

        return (ntext + " " + ctext, cnum * nnum)

    def ho_result(self, result, value):
        index = result.find('ho')
        count_ho = result.count('ho')

        if index != -1 and count_ho >= 1:
            index_rihun = result.find('rihun')
            value_str = len(str(value))
            if index_rihun != -1 and value_str > 7:
                result = result.replace("rihun ho", "ho rihun")
            lows = ["kuatr", "tr", "b", "m"]
            MEGA_SUFFIX = "iliaun"
            for low in lows:
                result = result.replace(
                    low + MEGA_SUFFIX + " ho", "ho " + low + MEGA_SUFFIX)
            remove_first_ho = result.startswith('ho')
            if remove_first_ho:
                result = result[3:]
        return result

    def remove_ho(self, result, value):
        value_str = str(value)
        result = self.ho_result(result, value)
        end_value = value_str[:-4]
        end_true = end_value.endswith('0')
        if end_true is False:
            if value > 100:
                if value_str[-1] != '0' and value_str[-2] == '0':
                    result = result.replace("ho", "")
                    result = result.replace("  ", " ")

        return result

    def to_cardinal(self, value):
        result = super().to_cardinal(value)

        results = self.remove_ho(result, value)
        return results

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        out = ""
        val = self.splitnum(value)
        outs = val
        while len(val) != 1:
            outs = []
            left, right = val[:2]
            if isinstance(left, tuple) and isinstance(right, tuple):
                outs.append(self.merge(left, right))
                if val[2:]:
                    outs.append(val[2:])
            else:
                for elem in val:
                    if isinstance(elem, list):
                        if len(elem) == 1:
                            outs.append(elem[0])
                        else:
                            outs.append(self.clean(elem))
                    else:
                        outs.append(elem)
            val = outs

        words, num = outs[0]

        words = self.remove_ho(words, value)

        if num in [90, 80, 70, 60, 50, 40, 30, 20, 10, 9, 8, 7, 5, 3, 2]:
            words = 'da'+words+'k'
        if num in [6, 4]:
            words = 'da'+words
        if num == 1:
            words = 'dahuluk'
        if num in [900, 800, 700, 500, 300, 200, 100]:
            words = 'dah'+words+'k'
        if num in [600, 400]:
            words = 'dah'+words

        words_split = words.split()
        if len(words_split) >= 3 and num < 100:
            first_word = 'da'+words_split[0]
            second_word = " ".join(words_split[1:])
            if 'haat' in second_word or 'neen' in second_word:
                words = first_word+" "+second_word
            else:
                words = first_word+" "+second_word+'k'

        word_first = 'dah'+words_split[0]
        if word_first == 'dahatus' and len(words_split) >= 3:
            word_second = " ".join(words_split[1:])
            if 'haat' in word_second or 'neen' in word_second:
                words = word_first+" "+word_second
            else:
                words = word_first+" "+word_second+'k'

        if len(str(num)) > 3:
            if 'haat' in words_split[-1:] or 'neen' in words_split[-1:]:
                words = 'da'+words
            else:
                words = 'da'+words+'k'

        result = self.title(out + words)

        return result

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%sº" % (value)

    def to_year(self, val, longval=True):
        if val < 0:
            return self.to_cardinal(abs(val)) + ' antes Kristu'
        return self.to_cardinal(val)

    def to_currency(self, val, currency='USD', cents=True):
        """
        Args:
            val: Numeric value
            currency (str): Currency code
            cents (bool): Verbose cents
            adjective (bool): Prefix currency name with adjective
        Returns:
            str: Formatted string

        """
        left, right, is_negative = parse_currency_parts(val)

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]

        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"' %
                (currency, self.__class__.__name__))

        minus_str = "%s " % self.negword.strip() if is_negative else ""
        money_str = self._money_verbose(left, currency)
        cents_str = self._cents_verbose(right, currency) \
            if cents else self._cents_terse(right, currency)

        if right == 0:
            return u'%s%s %s' % (
                minus_str,
                self.pluralize(left, cr1),
                money_str
            )
        else:
            return u'%s%s %s %s %s' % (
                minus_str,
                self.pluralize(left, cr1),
                money_str,
                self.pluralize(right, cr2),
                cents_str
            )



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from .base import Num2Word_Base
from .compat import strtype, to_s
from .currency import parse_currency_parts, prefix_currency


def select_text(text, reading=False, prefer=None):
    """Select the correct text from the Japanese number, reading and
    alternatives"""
    # select kanji number or kana reading
    if reading:
        text = text[1]
    else:
        text = text[0]

    # select the preferred one or the first one from multiple alternatives
    if not isinstance(text, strtype):
        common = set(text) & set(prefer or set())
        if len(common) == 1:
            text = common.pop()
        else:
            text = text[0]

    return text


def rendaku_merge_pairs(lpair, rpair):
    """Merge lpair < rpair while applying semi-irregular rendaku rules"""
    ltext, lnum = lpair
    rtext, rnum = rpair
    if lnum > rnum:
        raise ValueError

    if rpair == ("ひゃく", 100):
        if lpair == ("さん", 3):
            rtext = "びゃく"
        elif lpair == ("ろく", 6):
            ltext = "ろっ"
            rtext = "ぴゃく"
        elif lpair == ("はち", 8):
            ltext = "はっ"
            rtext = "ぴゃく"
    elif rpair == ("せん", 1000):
        if lpair == ("さん", 3):
            rtext = "ぜん"
        elif lpair == ("はち", 8):
            ltext = "はっ"
    elif rpair == ("ちょう", 10**12):
        if lpair == ("いち", 1):
            ltext = "いっ"
        elif lpair == ("はち", 8):
            ltext = "はっ"
        elif lpair == ("じゅう", 10):
            ltext = "じゅっ"
    elif rpair == ("けい", 10**16):
        if lpair == ("いち", 1):
            ltext = "いっ"
        elif lpair == ("ろく", 6):
            ltext = "ろっ"
        elif lpair == ("はち", 8):
            ltext = "はっ"
        elif lpair == ("じゅう", 10):
            ltext = "じゅっ"
        elif lpair == ("ひゃく", 100):
            ltext = "ひゃっ"

    return ("%s%s" % (ltext, rtext), lnum * rnum)


# Source: https://www.sljfaq.org/afaq/era-list.html
# if there are multiple eras for the same year, use the last one
ERA_START = [
    (645, ("大化", "たいか")),
    (650, ("白雉", "はくち")),
    (686, ("朱鳥", "しゅちょう")),
    (701, ("大宝", "たいほう")),
    (704, ("慶雲", "けいうん")),
    (708, ("和銅", "わどう")),
    (715, ("霊亀", "れいき")),
    (717, ("養老", "ようろう")),
    (724, ("神亀", "じんき")),
    (729, ("天平", "てんぴょう")),
    (749, ("天平感宝", "てんぴょうかんぽう")),
    (749, ("天平勝宝", "てんぴょうしょうほう")),
    (757, ("天平宝字", "てんぴょうじょうじ")),
    (765, ("天平神護", "てんぴょうじんご")),
    (767, ("神護景雲", "じんごけいうん")),
    (770, ("宝亀", "ほうき")),
    (781, ("天応", "てんおう")),
    (782, ("延暦", "えんりゃく")),
    (806, ("大同", "だいどう")),
    (810, ("弘仁", "こうにん")),
    (823, ("天長", "てんちょう")),
    (834, ("承和", "じょうわ")),
    (848, ("嘉祥", "かしょう")),
    (851, ("仁寿", "にんじゅ")),
    (855, ("斉衡", "さいこう")),
    (857, ("天安", "てんあん")),
    (859, ("貞観", "じょうがん")),
    (877, ("元慶", "がんぎょう")),
    (885, ("仁和", "にんな")),
    (889, ("寛平", "かんぴょう")),
    (898, ("昌泰", "しょうたい")),
    (901, ("延喜", "えんぎ")),
    (923, ("延長", "えんちょう")),
    (931, ("承平", "じょうへい")),
    (938, ("天慶", "てんぎょう")),
    (947, ("天暦", "てんりゃく")),
    (957, ("天徳", "てんとく")),
    (961, ("応和", "おうわ")),
    (964, ("康保", "こうほう")),
    (968, ("安和", "あんな")),
    (970, ("天禄", "てんろく")),
    (974, ("天延", "てんえん")),
    (976, ("貞元", "じょうげん")),
    (979, ("天元", "てんげん")),
    (983, ("永観", "えいかん")),
    (985, ("寛和", "かんな")),
    (987, ("永延", "えいえん")),
    (989, ("永祚", "えいそ")),
    (990, ("正暦", "しょうりゃく")),
    (995, ("長徳", "ちょうとく")),
    (999, ("長保", "ちょうほう")),
    (1004, ("寛弘", "かんこう")),
    (1013, ("長和", "ちょうわ")),
    (1017, ("寛仁", "かんにん")),
    (1021, ("治安", "じあん")),
    (1024, ("万寿", "まんじゅ")),
    (1028, ("長元", "ちょうげん")),
    (1037, ("長暦", "ちょうりゃく")),
    (1040, ("長久", "ちょうきゅう")),
    (1045, ("寛徳", "かんとく")),
    (1046, ("永承", "えいしょう")),
    (1053, ("天喜", "てんぎ")),
    (1058, ("康平", "こうへい")),
    (1065, ("治暦", "じりゃく")),
    (1069, ("延久", "えんきゅう")),
    (1074, ("承保", "じょうほう")),
    (1078, ("承暦", "じょうりゃく")),
    (1081, ("永保", "えいほう")),
    (1084, ("応徳", "おうとく")),
    (1087, ("寛治", "かんじ")),
    (1095, ("嘉保", "かほう")),
    (1097, ("永長", "えいちょう")),
    (1098, ("承徳", "じょうとく")),
    (1099, ("康和", "こうわ")),
    (1104, ("長治", "ちょうじ")),
    (1106, ("嘉承", "かじょう")),
    (1108, ("天仁", "てんにん")),
    (1110, ("天永", "てんねい")),
    (1113, ("永久", "えいきゅう")),
    (1118, ("元永", "げんえい")),
    (1120, ("保安", "ほうあん")),
    (1124, ("天治", "てんじ")),
    (1126, ("大治", "だいじ")),
    (1131, ("天承", "てんしょう")),
    (1132, ("長承", "ちょうしょう")),
    (1135, ("保延", "ほうえん")),
    (1141, ("永治", "えいじ")),
    (1142, ("康治", "こうじ")),
    (1144, ("天養", "てんよう")),
    (1145, ("久安", "きゅうあん")),
    (1151, ("仁平", "にんぺい")),
    (1154, ("久寿", "きゅうじゅ")),
    (1156, ("保元", "ほうげん")),
    (1159, ("平治", "へいじ")),
    (1160, ("永暦", "えいりゃく")),
    (1161, ("応保", "おうほう")),
    (1163, ("長寛", "ちょうかん")),
    (1165, ("永万", "えいまん")),
    (1166, ("仁安", "にんあん")),
    (1169, ("嘉応", "かおう")),
    (1171, ("承安", "しょうあん")),
    (1175, ("安元", "あんげん")),
    (1177, ("治承", "じしょう")),
    (1181, ("養和", "ようわ")),
    (1182, ("寿永", "じゅえい")),
    (1184, ("元暦", "げんりゃく")),
    (1185, ("文治", "ぶんじ")),
    (1190, ("建久", "けんきゅう")),
    (1199, ("正治", "しょうじ")),
    (1201, ("建仁", "けんにん")),
    (1204, ("元久", "げんきゅう")),
    (1206, ("建永", "けんえい")),
    (1207, ("承元", "じょうげん")),
    (1211, ("建暦", "けんりゃく")),
    (1214, ("建保", "けんぽう")),
    (1219, ("承久", "じょうきゅう")),
    (1222, ("貞応", "じょうおう")),
    (1225, ("元仁", "げんにん")),
    (1225, ("嘉禄", "かろく")),
    (1228, ("安貞", "あんてい")),
    (1229, ("寛喜", "かんき")),
    (1232, ("貞永", "じょうえい")),
    (1233, ("天福", "てんぷく")),
    (1235, ("文暦", "ぶんりゃく")),
    (1235, ("嘉禎", "かてい")),
    (1239, ("暦仁", "りゃくにん")),
    (1239, ("延応", "えんおう")),
    (1240, ("仁治", "にんじ")),
    (1243, ("寛元", "かんげん")),
    (1247, ("宝治", "ほうじ")),
    (1249, ("建長", "けんちょう")),
    (1256, ("康元", "こうげん")),
    (1257, ("正嘉", "しょうか")),
    (1259, ("正元", "しょうげん")),
    (1260, ("文応", "ぶんおう")),
    (1261, ("弘長", "こうちょう")),
    (1264, ("文永", "ぶんえい")),
    (1275, ("健治", "けんじ")),
    (1278, ("弘安", "こうあん")),
    (1288, ("正応", "しょうおう")),
    (1293, ("永仁", "えいにん")),
    (1299, ("正安", "しょうあん")),
    (1303, ("乾元", "けんげん")),
    (1303, ("嘉元", "かげん")),
    (1307, ("徳治", "とくじ")),
    (1308, ("延慶", "えんきょう")),
    (1311, ("応長", "おうちょう")),
    (1312, ("正和", "しょうわ")),
    (1317, ("文保", "ぶんぽう")),
    (1319, ("元応", "げんおう")),
    (1321, ("元亨", "げんこう")),
    (1325, ("正中", "しょうちゅ")),
    (1326, ("嘉暦", "かりゃく")),
    (1329, ("元徳", "げんとく")),
    (1331, ("元弘", "げんこう")),
    (1332, ("正慶", "しょうけい")),
    (1334, ("建武", "けんむ")),
    (1336, ("延元", "えいげん")),
    (1338, ("暦応", "りゃくおう")),
    (1340, ("興国", "こうこく")),
    (1342, ("康永", "こうえい")),
    (1345, ("貞和", "じょうわ")),
    (1347, ("正平", "しょうへい")),
    (1350, ("観応", "かんおう")),
    (1352, ("文和", "ぶんな")),
    (1356, ("延文", "えんぶん")),
    (1361, ("康安", "こうあん")),
    (1362, ("貞治", "じょうじ")),
    (1368, ("応安", "おうあん")),
    (1370, ("建徳", "けんとく")),
    (1372, ("文中", "ぶんちゅう")),
    (1375, ("永和", "えいわ")),
    (1375, ("天授", "てんじゅ")),
    (1379, ("康暦", "こうりゃく")),
    (1381, ("永徳", "えいとく")),
    (1381, ("弘和", "こうわ")),
    (1384, ("至徳", "しとく")),
    (1384, ("元中", "げんちゅう")),
    (1387, ("嘉慶", "かけい")),
    (1389, ("康応", "こうおう")),
    (1390, ("明徳", "めいとく")),
    (1394, ("応永", "おうえい")),
    (1428, ("正長", "しょうちょう")),
    (1429, ("永享", "えいきょう")),
    (1441, ("嘉吉", "かきつ")),
    (1444, ("文安", "ぶんあん")),
    (1449, ("宝徳", "ほうとく")),
    (1452, ("享徳", "きょうとく")),
    (1455, ("康正", "こうしょう")),
    (1457, ("長禄", "ちょうろく")),
    (1461, ("寛正", "かんしょう")),
    (1466, ("文正", "ぶんしょう")),
    (1467, ("応仁", "おうにん")),
    (1469, ("文明", "ぶんめい")),
    (1487, ("長享", "ちょうきょう")),
    (1489, ("延徳", "えんとく")),
    (1492, ("明応", "めいおう")),
    (1501, ("文亀", "ぶんき")),
    (1504, ("永正", "えいしょう")),
    (1521, ("大永", "だいえい")),
    (1528, ("享禄", "きょうろく")),
    (1532, ("天文", "てんぶん")),
    (1555, ("弘治", "こうじ")),
    (1558, ("永禄", "えいろく")),
    (1570, ("元亀", "げんき")),
    (1573, ("天正", "てんしょう")),
    (1593, ("文禄", "ぶんろく")),
    (1596, ("慶長", "けいちょう")),
    (1615, ("元和", "げんな")),
    (1624, ("寛永", "かんえい")),
    (1645, ("正保", "しょうほう")),
    (1648, ("慶安", "けいあん")),
    (1652, ("承応", "じょうおう")),
    (1655, ("明暦", "めいれき")),
    (1658, ("万治", "まんじ")),
    (1661, ("寛文", "かんぶん")),
    (1673, ("延宝", "えんぽう")),
    (1681, ("天和", "てんな")),
    (1684, ("貞享", "じょうきょう")),
    (1688, ("元禄", "げんろく")),
    (1704, ("宝永", "ほうえい")),
    (1711, ("正徳", "しょうとく")),
    (1716, ("享保", "きょうほう")),
    (1736, ("元文", "げんぶん")),
    (1741, ("寛保", "かんぽう")),
    (1744, ("延享", "えんきょう")),
    (1748, ("寛延", "かんえん")),
    (1751, ("宝暦", "ほうれき")),
    (1764, ("明和", "めいわ")),
    (1773, ("安永", "あんえい")),
    (1781, ("天明", "てんめい")),
    (1801, ("寛政", "かんせい")),
    (1802, ("享和", "きょうわ")),
    (1804, ("文化", "ぶんか")),
    (1818, ("文政", "ぶんせい")),
    (1831, ("天保", "てんぽう")),
    (1845, ("弘化", "こうか")),
    (1848, ("嘉永", "かえい")),
    (1855, ("安政", "あんせい")),
    (1860, ("万延", "まんえい")),
    (1861, ("文久", "ぶんきゅう")),
    (1864, ("元治", "げんじ")),
    (1865, ("慶応", "けいおう")),
    (1868, ("明治", "めいじ")),
    (1912, ("大正", "たいしょう")),
    (1926, ("昭和", "しょうわ")),
    (1989, ("平成", "へいせい")),
    (2019, ("令和", "れいわ")),
]


class Num2Word_JA(Num2Word_Base):
    CURRENCY_FORMS = {
        'JPY': (('円', 'えん'), ()),
    }

    def set_high_numwords(self, high):
        max = 4 * len(high)
        for word, n in zip(high, range(max, 0, -4)):
            self.cards[10 ** n] = word

    def setup(self):
        self.negword = "マイナス"
        self.pointword = ("点", "てん")
        self.exclude_title = ["点", "マイナス"]

        self.high_numwords = [
            ("万", "まん"),    # 10**4 man
            ("億", "おく"),    # 10**8 oku
            ("兆", "ちょう"),  # 10**12 chō
            ("京", "けい"),    # 10**16 kei
            ("垓", "がい"),    # 10**20 gai
            ("秭", "し"),      # 10**24 shi
            ("穣", "じょう"),  # 10**28 jō
            ("溝", "こう"),    # 10**32 kō
            ("澗", "かん"),    # 10**36 kan
            ("正", "せい"),    # 10**40 sei
            ("載", "さい"),    # 10**44 sai
            ("極", "ごく"),    # 10**48 goku
        ]

        self.high_numwords.reverse()

        self.mid_numwords = [
            (1000, ("千", "せん")),
            (100, ("百", "ひゃく")),
        ]

        self.low_numwords = [
            ("十", "じゅう"),                  # 10 jū
            ("九", "きゅう"),                  # 9 kyū
            ("八", "はち"),                    # 8 hachi
            ("七", ("なな", "しち")),          # 7 nana, shichi
            ("六", "ろく"),                    # 6 roku
            ("五", "ご"),                      # 5 go
            ("四", ("よん", "し")),            # 4 yon, shi
            ("三", "さん"),                    # 3 san
            ("二", "に"),                      # 2 ni
            ("一", "いち"),                    # 1 ichi
            # both are alternatives, 零 doesn't map to ゼロ or 〇 to れい
            (("零", "〇"), ("ゼロ", "れい")),  # 0 ZERO, rei
        ]

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair

        fmt = "%s%s"
        # ignore lpair if lnum is 1 and rnum is less than 10000
        if lnum == 1 and rnum < 10000:
            return rpair
        # rnum is added to lnum
        elif lnum > rnum:
            return (fmt % (ltext, rtext), lnum + rnum)
        # rnum is multiplied by lnum
        elif lnum < rnum:
            return rendaku_merge_pairs(lpair, rpair)

    def _ordinal_suffix(self, reading, counter):
        if reading:
            if counter == "番":
                return "ばんめ"
            else:
                raise NotImplementedError(
                    "Reading not implemented for %s" % counter)
        else:
            return counter + "目"

    def to_ordinal(self, value, reading=False, prefer=None, counter="番"):
        self.verify_ordinal(value)
        base = self.to_cardinal(value, reading=reading, prefer=prefer)
        return "%s%s" % (base, self._ordinal_suffix(reading, counter))

    def to_ordinal_num(self, value, reading=False, counter="番"):
        return "%s%s" % (value, self._ordinal_suffix(reading, counter))

    def to_year(self, val, suffix=None, longval=True, reading=False,
                prefer=None, era=True):
        year = val
        # Gregorian calendar
        if not era:
            prefix = ""
            if year < 0:
                year = abs(year)
                prefix = "きげんぜん" if reading else "紀元前"

            year_words = self.to_cardinal(year, reading=reading, prefer=prefer)
            if reading and year % 10 == 9:
                year_words = year_words[:-3] + "く"

            return "%s%s%s" % (prefix, year_words, "ねん" if reading else "年")

        # Era calendar (default)
        min_year = ERA_START[0][0]
        last_era_idx = len(ERA_START) - 1
        if year < min_year:
            raise ValueError(
                "Can't convert years less than %s to era" % min_year)

        first = 0
        last = last_era_idx
        era_idx = None
        while era_idx is None:
            mid = (first + last) // 2
            if mid == last_era_idx or (ERA_START[mid][0] <= year and
                                       ERA_START[mid + 1][0] > year):
                era_idx = mid
                # if an era lasting less than a year is preferred, choose it
                if prefer:
                    i = mid - 1
                    while i >= 0 and ERA_START[i][0] == year:
                        # match kanji or hiragana
                        if set(ERA_START[i][1]) & set(prefer):
                            era_idx = i
                            break
                        i -= 1

            # ends up at the last index where year >= ERA_START[mid][0]
            if year < ERA_START[mid][0]:
                last = mid - 1
            else:
                first = mid + 1

        era = ERA_START[era_idx]
        era_name = era[1][0]
        era_year = year - era[0] + 1
        fmt = "%s%s年"
        if reading == "arabic":
            era_year_words = str(era_year)
        elif reading:
            era_name = era[1][1]
            era_year_words = (self.to_cardinal(era_year, reading=True,
                                               prefer=prefer)
                              if era_year != 1 else "がん")
            if era_year % 10 == 9:
                era_year_words = era_year_words[:-3] + "く"
            fmt = "%s%sねん"
        else:
            era_year_words = (self.to_cardinal(era_year, reading=False,
                                               prefer=prefer)
                              if era_year != 1 else "元")

        return fmt % (era_name, era_year_words)

    def to_currency(self, val, currency="JPY", cents=False, separator="",
                    adjective=False, reading=False, prefer=None):
        left, right, is_negative = parse_currency_parts(
            val, is_int_with_cents=cents)

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]
            if (cents or abs(val) != left) and not cr2:
                raise ValueError('Decimals not supported for "%s"' % currency)
        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"' %
                (currency, self.__class__.__name__))

        if adjective and currency in self.CURRENCY_ADJECTIVES:
            cr1 = prefix_currency(self.CURRENCY_ADJECTIVES[currency], cr1)

        minus_str = self.negword if is_negative else ""

        return '%s%s%s%s%s' % (
            minus_str,
            self.to_cardinal(left, reading=reading, prefer=prefer),
            cr1[1] if reading else cr1[0],
            self.to_cardinal(right, reading=reading, prefer=prefer)
            if cr2 else '',
            (cr2[1] if reading else cr2[0]) if cr2 else '',
        )

    def splitnum(self, value, reading, prefer):
        for elem in self.cards:
            if elem > value:
                continue

            out = []
            if value == 0:
                div, mod = 1, 0
            else:
                div, mod = divmod(value, elem)

            if div == 1:
                out.append((select_text(self.cards[1], reading, prefer), 1))
            else:
                if div == value:  # The system tallies, eg Roman Numerals
                    return [(
                        div * select_text(self.cards[elem], reading, prefer),
                        div * elem)]
                out.append(self.splitnum(div, reading, prefer))

            out.append((select_text(self.cards[elem], reading, prefer), elem))

            if mod:
                out.append(self.splitnum(mod, reading, prefer))

            return out

    def to_cardinal(self, value, reading=False, prefer=None):
        try:
            assert int(value) == value
        except (ValueError, TypeError, AssertionError):
            return self.to_cardinal_float(value, reading=reading,
                                          prefer=prefer)

        out = ""
        if value < 0:
            value = abs(value)
            out = self.negword

        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        val = self.splitnum(value, reading, prefer)
        words, _ = self.clean(val)
        return self.title(out + words)

    def to_cardinal_float(self, value, reading=False, prefer=None):
        prefer = prefer or ["れい"]
        try:
            float(value) == value
        except (ValueError, TypeError, AssertionError):
            raise TypeError(self.errmsg_nonnum % value)

        pre, post = self.float2tuple(float(value))

        post = str(post)
        post = '0' * (self.precision - len(post)) + post

        out = [self.to_cardinal(pre, reading=reading, prefer=prefer)]
        if self.precision:
            out.append(self.title(self.pointword[1 if reading else 0]))

        for i in range(self.precision):
            curr = int(post[i])
            out.append(to_s(
                self.to_cardinal(curr, reading=reading, prefer=prefer)))

        return "".join(out)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from . import (lang_AM, lang_AR, lang_AZ, lang_BE, lang_BN, lang_CA, lang_CE,
               lang_CS, lang_CY, lang_DA, lang_DE, lang_EN, lang_EN_IN,
               lang_EN_NG, lang_EO, lang_ES, lang_ES_CO, lang_ES_CR,
               lang_ES_GT, lang_ES_NI, lang_ES_VE, lang_FA, lang_FI, lang_FR,
               lang_FR_BE, lang_FR_CH, lang_FR_DZ, lang_HE, lang_HU, lang_ID,
               lang_IS, lang_IT, lang_JA, lang_KN, lang_KO, lang_KZ, lang_LT,
               lang_LV, lang_NL, lang_NO, lang_PL, lang_PT, lang_PT_BR,
               lang_RO, lang_RU, lang_SK, lang_SL, lang_SR, lang_SV, lang_TE,
               lang_TET, lang_TG, lang_TH, lang_TR, lang_UK, lang_VI)

CONVERTER_CLASSES = {
    'am': lang_AM.Num2Word_AM(),
    'ar': lang_AR.Num2Word_AR(),
    'az': lang_AZ.Num2Word_AZ(),
    'be': lang_BE.Num2Word_BE(),
    'bn': lang_BN.Num2Word_BN(),
    'ca': lang_CA.Num2Word_CA(),
    'ce': lang_CE.Num2Word_CE(),
    'cs': lang_CS.Num2Word_CS(),
    'cy': lang_CY.Num2Word_CY(),
    'en': lang_EN.Num2Word_EN(),
    'en_IN': lang_EN_IN.Num2Word_EN_IN(),
    'en_NG': lang_EN_NG.Num2Word_EN_NG(),
    'fa': lang_FA.Num2Word_FA(),
    'fr': lang_FR.Num2Word_FR(),
    'fr_CH': lang_FR_CH.Num2Word_FR_CH(),
    'fr_BE': lang_FR_BE.Num2Word_FR_BE(),
    'fr_DZ': lang_FR_DZ.Num2Word_FR_DZ(),
    'de': lang_DE.Num2Word_DE(),
    'fi': lang_FI.Num2Word_FI(),
    'eo': lang_EO.Num2Word_EO(),
    'es': lang_ES.Num2Word_ES(),
    'es_CO': lang_ES_CO.Num2Word_ES_CO(),
    'es_CR': lang_ES_CR.Num2Word_ES_CR(),
    'es_GT': lang_ES_GT.Num2Word_ES_GT(),
    'es_NI': lang_ES_NI.Num2Word_ES_NI(),
    'es_VE': lang_ES_VE.Num2Word_ES_VE(),
    'id': lang_ID.Num2Word_ID(),
    'ja': lang_JA.Num2Word_JA(),
    'kn': lang_KN.Num2Word_KN(),
    'ko': lang_KO.Num2Word_KO(),
    'kz': lang_KZ.Num2Word_KZ(),
    'lt': lang_LT.Num2Word_LT(),
    'lv': lang_LV.Num2Word_LV(),
    'pl': lang_PL.Num2Word_PL(),
    'ro': lang_RO.Num2Word_RO(),
    'ru': lang_RU.Num2Word_RU(),
    'sk': lang_SK.Num2Word_SK(),
    'sl': lang_SL.Num2Word_SL(),
    'sr': lang_SR.Num2Word_SR(),
    'sv': lang_SV.Num2Word_SV(),
    'no': lang_NO.Num2Word_NO(),
    'da': lang_DA.Num2Word_DA(),
    'pt': lang_PT.Num2Word_PT(),
    'pt_BR': lang_PT_BR.Num2Word_PT_BR(),
    'he': lang_HE.Num2Word_HE(),
    'it': lang_IT.Num2Word_IT(),
    'vi': lang_VI.Num2Word_VI(),
    'tg': lang_TG.Num2Word_TG(),
    'th': lang_TH.Num2Word_TH(),
    'tr': lang_TR.Num2Word_TR(),
    'nl': lang_NL.Num2Word_NL(),
    'uk': lang_UK.Num2Word_UK(),
    'te': lang_TE.Num2Word_TE(),
    'tet': lang_TET.Num2Word_TET(),
    'hu': lang_HU.Num2Word_HU(),
    'is': lang_IS.Num2Word_IS(),
}

CONVERTES_TYPES = ['cardinal', 'ordinal', 'ordinal_num', 'year', 'currency']


def num2words(number, ordinal=False, lang='en', to='cardinal', **kwargs):
    # We try the full language first
    if lang not in CONVERTER_CLASSES:
        # ... and then try only the first 2 letters
        lang = lang[:2]
    if lang not in CONVERTER_CLASSES:
        raise NotImplementedError()
    converter = CONVERTER_CLASSES[lang]

    if isinstance(number, str):
        number = converter.str_to_number(number)

    # backwards compatible
    if ordinal:
        to = 'ordinal'

    if to not in CONVERTES_TYPES:
        raise NotImplementedError()

    return getattr(converter, 'to_{}'.format(to))(number, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (c) 2023, Johannes Heinecke.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .currency import parse_currency_parts
from .lang_EU import Num2Word_EU

# Chechen numbers inflect in case if without noun or
# use a special oblique ending when followed by a counted noun
# 4, 14, 40 and composites thereof agree in class (gender) with the
# noun. Chechen has 6 classes which are indicated by the initial
# letter of 4, 14 and 40. By default it is "д" but
# it can also be "б", "й" or "в".
# Indicate the needed class prefix as follows
#   num2words(4, lang='ce', case="abs", clazz="б")


CARDINALS = {
    "casenames": {
        "abs": "Им.",
        "gen": "Род.",
        "dat": "Дат.",
        "erg": "Эрг;",
        "instr": "Твор.",
        "mat": "Вещ.",
        "comp": "Сравнит.",
        "all": "Местн.",
    },
    "casesuffix_cons": {  # to be added to numerals with final consonant
        "gen": "аннан",
        "dat": "анна",
        "erg": "амма",
        "instr": "анца",
        "mat": "аннах",
        "comp": "аннал",
        "all": "анга",
        "obl": "ан",
        "ORD": "алгӀа",
    },
    "casesuffix_voc": {  # to be added to numerals with final vowel
        "gen": "ннан",
        "dat": "нна",
        "erg": "мма",
        "instr": "нца",
        "mat": "ннах",
        "comp": "ннал",
        "all": "нга",
        "obl": "н",
        "ORD": "лгӀа",
    },
    0: {
        "attr": "ноль",
        "abs": "ноль",
        "gen": "нолан",
        "dat": "нолана",
        "erg": "ноло",
        "instr": "ноланца",
        "mat": "ноланах",
        "comp": "ноланал",
        "all": "ноланга",
    },
    1: {
        "attr": "цхьа",  # in front of nouns in ABS
        "obl": "цхьана",  # with nouns in other cases than ABS
        "abs": "цхьаъ",
        "gen": "цхьаннан",
        "dat": "цхьанна",
        "erg": "цхьамма",
        "instr": "цхьаьнца",
        "mat": "цхьаннах",
        "comp": "цхьаннал",
        "all": "цхаьнга",
        "ORD": "цхьалгӀа",
    },
    2: {
        "attr": "ши",  # in front of 100, 1000
        "obl": "шина",
        "abs": "шиъ",
        "gen": "шиннан",
        "dat": "шинна",
        "erg": "шимма",
        "instr": "шинца",
        "mat": "шиннах",
        "comp": "шиннал",
        "all": "шинга",
        "ORD": "шолгӀа",
    },
    3: {
        "attr": "кхо",
        "obl": "кхона",
        "abs": "кхоъ",
        "gen": "кхааннан",
        "dat": "кхаанна",
        "erg": "кхаамма",
        "instr": "кхаанца",
        "mat": "кхааннах",
        "comp": "кхааннал",
        "all": "кхаанга",
        "ORD": "кхоалгӀа",
    },
    4: {
        "attr": "д*и",
        "obl": "д*еа",
        "abs": "д*иъ",
        "gen": "д*еаннан",
        "dat": "д*еанна",
        "erg": "д*еамма",
        "instr": "д*еанца",
        "mat": "д*еаннах",
        "comp": "д*еаннал",
        "all": "д*еанга",
        "ORD": "д*оьалгӀа",
    },
    5: {
        "attr": "пхи",
        "obl": "пхеа",
        "abs": "пхиъ",
        "gen": "пхеаннан",
        "dat": "пхеанна",
        "erg": "пхеамма",
        "instr": "нхеанца",
        "mat": "пхеаннах",
        "comp": "пхеаннал",
        "all": "пхеанга",
        "ORD": "пхоьалгӀа",
    },
    6: {
        "abs": "ялх",
        "attr": "ялх",
        "ORD": "йолхалгӀа",
    },
    7: {
        "abs": "ворхӀ",
        "attr": "ворхӀ",
        "ORD": "ворхӀалгӀа",
    },
    8: {
        "abs": "бархӀ",
        "attr": "бархӀ",
        "ORD": "борхӀалӀа",
    },
    9: {
        "abs": "исс",
        "attr": "исс",
        "ORD": "уьссалгӀа",
    },
    10: {
        "attr": "итт",
        "abs": "итт",
        "gen": "иттаннан",
        "dat": "иттанна",
        "erg": "иттамма",
        "instr": "иттанца",
        "mat": "иттаннах",
        "comp": "иттаннал",
        "all": "иттанга",
        "ORD": "уьтталгӀа",
    },
    11: {
        "abs": "цхьайтта",
        "attr": "цхьайтта",
        "ORD": "цхьайтталгӀа",
    },
    12: {
        "abs": "шийтта",
        "attr": "шийтта",
        "ORD": "шийтталга",
    },
    13: {
        "abs": "кхойтта",
        "attr": "кхойтта",
        "ORD": "кхойтталгӀа",
    },
    14: {
        "abs": "д*ейтта",
        "attr": "д*ейтта",
        "ORD": "д*ейтталгӀа",
    },
    15: {
        "abs": "пхийтта",
        "attr": "пхийтта",
        "ORD": "пхийтталгӀа",
    },
    16: {
        "abs": "ялхитта",
        "attr": "ялхитта",
        "ORD": "ялхитталгӀа",
    },
    17: {
        "abs": "вуьрхӀитта",
        "attr": "вуьрхӀитта",
        "ORD": "вуьрхӀитталгӀа",
    },
    18: {
        "abs": "берхӀитта",
        "attr": "берхӀитта",
        "ORD": "берхитталӀа",
    },
    19: {
        "abs": "ткъайесна",
        "attr": "ткъайесна",
        "ORD": "ткъаесналгӀа",
    },
    20: {
        "abs": "ткъа",
        "gen": "ткъаннан",
        "dat": "ткъанна",
        "erg": "ткъамма",
        "instr": "ткъанца",
        "mat": "ткъаннах",
        "comp": "ткъаннал",
        "all": "ткъанга",
        "attr": "ткъе",
        "ORD": "ткъолгӀа",
    },
    40: {
        "abs": "шовзткъа",
        "attr": "шовзткъе",
        "ORD": "шовзткъалгІа",
    },
    60: {
        "abs": "кхузткъа",
        "attr": "кхузткъе",
        "ORD": "кхузткъалгІа",
    },
    80: {
        "abs": "дезткъа",
        "attr": "дезткъе",
        "ORD": "дезткъалгІа",
    },
    100: {
        "attr": "бӀе",
        "abs": "бӀе",
        "obl": "бӀен",
        "gen": "бӀеннан",
        "dat": "бӀенна",
        "erg": "бӀемма",
        "instr": "бӀенца",
        "mat": "бӀеннах",
        "comp": "бӀеннал",
        "all": "бӀенга",
        "ORD": "бІолгІа",
    },
    1000: {
        "attr": "эзар",
        "abs": "эзар",
        "obl": "эзаран",
        "gen": "эзарнан",
        "dat": "эзарна",
        "erg": "эзарно",
        "instr": "эзарнаца",
        "mat": "эзарнах",
        "comp": "эзарнал",
        "all": "эзаранга",
        "ORD": "эзарлагІа",
    },
    1000000: {
        "attr": "миллион",
        "abs": "миллион",
        "ORD": "миллионалгІа",
    },
}

ILLIONS = {
    6: {
        "attr": "миллион",
        "abs": "миллион",
        "ORD": "миллионалгІа",
    },
    9: {
        "attr": "миллиард",
        "abs": "миллиард",
        "ORD": "миллиардалгІа",
    },
    12: {
        "attr": "биллион",
        "abs": "биллион",
        "ORD": "биллионалгІа",
    },
    15: {
        "attr": "биллиард",
        "abs": "биллиард",
        "ORD": "биллиардалгІа",
    },
    18: {
        "attr": "триллион",
        "abs": "триллион",
        "ORD": "триллионалгІа",
    },
    21: {
        "attr": "триллиард",
        "abs": "триллиард",
        "ORD": "триллиардалгІа",
    },
    24: {
        "attr": "квадриллион",
        "abs": "квадриллион",
        "ORD": "квадриллионалгІа",
    },
    27: {
        "attr": "квадриллиард",
        "abs": "квадриллиард",
        "ORD": "квадриллиардалгІа",
    },
    30: {
        "attr": "квинтиллион",
        "abs": "квинтиллион",
        "ORD": "квинтиллионалгІа",
    },
    33: {
        "attr": "квинтиллиард",
        "abs": "квинтиллиард",
        "ORD": "квинтиллиардалгІа",
    },
}


MINUS = "минус"
# DECIMALPOINT = "запятая"  # check !
DECIMALPOINT = "а"


class Num2Word_CE(Num2Word_EU):
    CURRENCY_FORMS = {
        # currency code: (sg, pl), (sg, pl)
        "EUR": (("Евро", "Евро"), ("Сент", "Сенташ")),
        "RUB": (("Сом", "Сомаш"), ("Кепек", "Кепекаш")),
        "USD": (("Доллар", "Доллараш"), ("Сент", "Сенташ")),
        "GBP": (("Фунт", "Фунташ"), ("Пенни", "Пенни")),
    }

    def setup(self):
        Num2Word_EU.setup(self)
        self.negword = "минус"
        self.pointword = "запятая"  # check !
        # self.errmsg_nonnum = (
        #    u"Seulement des nombres peuvent être convertis en mots."
        #    )
        # self.errmsg_toobig = (
        #    u"Nombre trop grand pour être converti en mots (abs(%s) > %s)."
        #    )
        # self.exclude_title = ["et", "virgule", "moins"]
        self.mid_numwords = []
        self.low_numwords = []
        self.ords = {}

    def to_ordinal(self, number, clazz="д"):
        # implement here your code. number is the integer to
        # be transformed into an ordinal as a word (str)
        # which is returned
        return self.to_cardinal(number, clazz=clazz, case="ORD")

    def to_cardinal(self, number, clazz="д", case="abs"):
        if isinstance(number, float):
            entires = self.to_cardinal(int(number))
            float_part = str(number).split(".")[1]
            postfix = " ".join(
                # Drops the trailing zero and comma
                [self.to_cardinal(int(c)) for c in float_part]
            )
            return entires + " " + DECIMALPOINT + " " + postfix

        elif number < 20:
            return self.makecase(number, case, clazz)
        elif number < 100:
            twens = number // 20
            units = number % 20
            base = twens * 20
            if units == 0:
                return self.makecase(number, case, clazz)
            else:
                twenties = self.makecase(base, "attr", clazz)
                rest = self.to_cardinal(units, clazz=clazz, case=case)
                return twenties + " " + rest.replace("д*", clazz)
        elif number < 1000:
            hundreds = number // 100
            tens = number % 100
            if hundreds > 1:
                hundert = (
                    CARDINALS[hundreds]["attr"].replace("д*", clazz) + " "
                )
            else:
                hundert = ""
            if tens != 0:
                rest = self.to_cardinal(tens, clazz=clazz, case=case)
                return hundert + CARDINALS[100]["abs"] + " " + rest
            else:
                return hundert + self.makecase(100, case, clazz)
        elif number < 1000000:
            thousands = number // 1000
            hundert = number % 1000
            if hundert > 0:
                tcase = "attr"
            else:
                tcase = case
            if thousands > 1:
                tausend = (
                    self.to_cardinal(thousands, clazz=clazz, case="attr")
                    + " "
                    + CARDINALS[1000][tcase]
                )
            else:
                tausend = self.makecase(1000, tcase, clazz)

            if hundert != 0:
                rest = " " + self.to_cardinal(hundert, clazz=clazz, case=case)
            else:
                rest = ""
            return tausend + rest

        elif number < 10**34:
            out = []
            for pot in reversed([6, 9, 12, 15, 18, 21, 24, 27, 30, 33]):
                # 3 digits of billion, trillion etc
                step = number // 10**pot % 1000
                if step > 0:
                    words = self.to_cardinal(step, clazz=clazz, case="attr")
                    out.append(words + " " + ILLIONS[pot]["attr"])
            rest = number % 10**6
            if rest:
                out.append(self.to_cardinal(rest, clazz=clazz, case=case))
            return " ".join(out)

        return "NOT IMPLEMENTED"

    def _money_verbose(self, number, currency, case):
        mcase = "attr"
        if case != "abs":
            mcase = "obl"
        return self.to_cardinal(number, case=mcase)

    def _cents_verbose(self, number, currency, case):
        mcase = "attr"
        if case != "abs":
            mcase = "obl"
        return self.to_cardinal(number, case=mcase)

    def to_currency(
        self,
        val,
        currency="RUB",
        cents=True,
        separator=",",
        adjective=False,
        case="abs",
    ):
        """
        Args:
            val: Numeric value
            currency (str): Currency code
            cents (bool): Verbose cents
            separator (str): Cent separator
            adjective (bool): Prefix currency name with adjective
        Returns:
            str: Formatted string

        """
        left, right, is_negative = parse_currency_parts(val)

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]
            devise = cr1[0]
            centime = cr2[0]
        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"'
                % (currency, self.__class__.__name__)
            )

        minus_str = "%s " % self.negword.strip() if is_negative else ""
        money_str = self._money_verbose(left, currency, case)
        cents_str = (
            self._cents_verbose(right, currency, case)
            if cents
            else self._cents_terse(right, currency)
        )

        return "%s%s %s%s %s %s" % (
            minus_str,
            money_str,
            devise,  # always singular
            separator,
            cents_str,
            centime,
        )

    def to_ordinal_num(self, number):
        self.verify_ordinal(number)
        return str(number) + "-й"

    def to_year(self, year, case="abs"):
        return self.to_cardinal(year, case=case)

    def makecase(self, number, case, clazz):
        # print("ZZZZ", number, CARDINALS[number])
        if case in CARDINALS[number]:
            return CARDINALS[number][case].replace("д*", clazz)
        else:
            if CARDINALS[number]["abs"][-1] in "а":
                return (
                    CARDINALS[number]["abs"].replace("д*", clazz)
                    + CARDINALS["casesuffix_voc"][case]
                )
            else:
                return (
                    CARDINALS[number]["abs"].replace("д*", clazz)
                    + CARDINALS["casesuffix_cons"][case]
                )



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import print_function, unicode_literals

from .lang_EU import Num2Word_EU


class Num2Word_FR(Num2Word_EU):
    CURRENCY_FORMS = {
        'EUR': (('euro', 'euros'), ('centime', 'centimes')),
        'USD': (('dollar', 'dollars'), ('cent', 'cents')),
        'FRF': (('franc', 'francs'), ('centime', 'centimes')),
        'GBP': (('livre', 'livres'), ('penny', 'pence')),
        'CNY': (('yuan', 'yuans'), ('fen', 'jiaos')),
    }

    def setup(self):
        Num2Word_EU.setup(self)

        self.negword = "moins "
        self.pointword = "virgule"
        self.errmsg_nonnum = (
            u"Seulement des nombres peuvent être convertis en mots."
            )
        self.errmsg_toobig = (
            u"Nombre trop grand pour être converti en mots (abs(%s) > %s)."
            )
        self.exclude_title = ["et", "virgule", "moins"]
        self.mid_numwords = [(1000, "mille"), (100, "cent"),
                             (80, "quatre-vingts"), (60, "soixante"),
                             (50, "cinquante"), (40, "quarante"),
                             (30, "trente")]
        self.low_numwords = ["vingt", "dix-neuf", "dix-huit", "dix-sept",
                             "seize", "quinze", "quatorze", "treize", "douze",
                             "onze", "dix", "neuf", "huit", "sept", "six",
                             "cinq", "quatre", "trois", "deux", "un", "zéro"]
        self.ords = {
            "cinq": "cinquième",
            "neuf": "neuvième",
        }

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next

        if cnum == 1:
            if nnum < 1000000:
                return next
        else:
            if (not (cnum - 80) % 100
                or (not cnum % 100 and cnum < 1000))\
                    and nnum < 1000000 \
                    and ctext[-1] == "s":
                ctext = ctext[:-1]
            if cnum < 1000 and nnum != 1000 and \
                    ntext[-1] != "s" and not nnum % 100:
                ntext += "s"

        if nnum < cnum < 100:
            if nnum % 10 == 1 and cnum != 80:
                return ("%s et %s" % (ctext, ntext), cnum + nnum)
            return ("%s-%s" % (ctext, ntext), cnum + nnum)
        if nnum > cnum:
            return ("%s %s" % (ctext, ntext), cnum * nnum)
        return ("%s %s" % (ctext, ntext), cnum + nnum)

    # Is this right for such things as 1001 - "mille unième" instead of
    # "mille premier"??  "millième"??

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        if value == 1:
            return "premier"
        word = self.to_cardinal(value)
        for src, repl in self.ords.items():
            if word.endswith(src):
                word = word[:-len(src)] + repl
                break
        else:
            if word[-1] == "e":
                word = word[:-1]
            word = word + "ième"
        return word

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        out = str(value)
        out += "er" if value == 1 else "me"
        return out

    def to_currency(self, val, currency='EUR', cents=True, separator=' et',
                    adjective=False):
        result = super(Num2Word_FR, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)
        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.
# Copyright (c) 2018, Abdullah Alhazmy, Alhazmy13.  All Rights Reserved.
# Copyright (c) 2020, Hamidreza Kalbasi.  All Rights Reserved.
# Copyright (c) 2023, Nika Soltani Tehrani.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from decimal import Decimal
from math import floor

farsiOnes = [
    "", "یک", "دو", "سه", "چهار", "پنج", "شش", "هفت", "هشت",
    "نه",
    "ده",
    "یازده",
    "دوازده",
    "سیزده",
    "چهارده",
    "پانزده",
    "شانزده",
    "هفده",
    "هجده",
    "نوزده",
]

farsiTens = [
    "",
    "ده",
    "بیست",
    "سی",
    "چهل",
    "پنجاه",
    "شصت",
    "هفتاد",
    "هشتاد",
    "نود",
]

farsiHundreds = [
    "",
    "صد",
    "دویست",
    "سیصد",
    "چهارصد",
    "پانصد",
    "ششصد",
    "هفتصد",
    "هشتصد",
    "نهصد",
]

farsiBig = [
    '',
    ' هزار',
    ' میلیون',
    " میلیارد",
    ' تریلیون',
    " تریلیارد",
]

farsiFrac = ["", "دهم", "صدم"]
farsiFracBig = ["", "هزارم", "میلیونیم", "میلیاردیم"]

farsiSeperator = ' و '


class Num2Word_FA(object):
    # Those are unused
    errmsg_toobig = "Too large"
    MAXNUM = 10 ** 36

    def __init__(self):
        self.number = 0

    def float2tuple(self, value):
        pre = int(value)

        # Simple way of finding decimal places to update the precision
        self.precision = abs(Decimal(str(value)).as_tuple().exponent)

        post = abs(value - pre) * 10**self.precision
        if abs(round(post) - post) < 0.01:
            # We generally floor all values beyond our precision (rather than
            # rounding), but in cases where we have something like 1.239999999,
            # which is probably due to python's handling of floats, we actually
            # want to consider it as 1.24 instead of 1.23
            post = int(round(post))
        else:
            post = int(floor(post))
        return pre, post, self.precision

    def cardinal3(self, number):
        if number <= 19:
            return farsiOnes[number]
        if number < 100:
            x, y = divmod(number, 10)
            if y == 0:
                return farsiTens[x]
            return farsiTens[x] + farsiSeperator + farsiOnes[y]
        x, y = divmod(number, 100)
        if y == 0:
            return farsiHundreds[x]
        return farsiHundreds[x] + farsiSeperator + self.cardinal3(y)

    def cardinalPos(self, number):
        x = number
        res = ''
        for b in farsiBig:
            x, y = divmod(x, 1000)
            if y == 0:
                continue
            yx = self.cardinal3(y) + b
            if b == ' هزار' and y == 1:
                yx = 'هزار'
            if res == '':
                res = yx
            else:
                res = yx + farsiSeperator + res
        return res

    def fractional(self, number, level):
        if number == 5:
            return "نیم"
        x = self.cardinalPos(number)
        ld3, lm3 = divmod(level, 3)
        ltext = (farsiFrac[lm3] + " " + farsiFracBig[ld3]).strip()
        return x + " " + ltext

    def to_currency(self, value):
        return self.to_cardinal(value) + " تومان"

    def to_ordinal(self, number):
        r = self.to_cardinal(number)
        if r[-1] == 'ه' and r[-2] == 'س':
            return r[:-1] + 'وم'
        return r + 'م'

    def to_year(self, value):
        return self.to_cardinal(value)

    @staticmethod
    def to_ordinal_num(value):
        return str(value)+"م"

    def to_cardinal(self, number):
        if number < 0:
            return "منفی " + self.to_cardinal(-number)
        if number == 0:
            return "صفر"
        x, y, level = self.float2tuple(number)
        if y == 0:
            return self.cardinalPos(x)
        if x == 0:
            return self.fractional(y, level)
        return self.cardinalPos(x) + farsiSeperator + self.fractional(y, level)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base
from .utils import get_digits, splitbyx

GENDER_PLURAL_INDEXES = {
    'm': 0, 'masculine': 0, 'м': 0, 'мужской': 0,
    'f': 1, 'feminine': 1, 'ж': 0, 'женский': 0,
    'n': 2, 'neuter': 2, 'с': 0, 'средний': 0,
    'p': 3, 'plural': 3
}
CASE_INDEXES = {
    'n': 0, 'nominative': 0, 'и': 0, 'именительный': 0,
    'g': 1, 'genitive': 1, 'р': 1, 'родительный': 1,
    'd': 2, 'dative': 2, 'д': 2, 'дательный': 2,
    'a': 3, 'accusative': 3, 'в': 3, 'винительный': 3,
    'i': 4, 'instrumental': 4, 'т': 4, 'творительный': 4,
    'p': 5, 'prepositional': 5, 'п': 5, 'предложный': 5
}
# Default values
D_CASE = 'n'
D_PLURAL = False
D_GENDER = 'm'
D_ANIMATE = True


def get_num_element(cases_dict, num, **kwargs):
    return case_classifier_element(cases_dict[num], **kwargs)


def case_classifier_element(classifier, case=D_CASE, plural=D_PLURAL,
                            gender=D_GENDER, animate=D_ANIMATE):
    case = classifier[CASE_INDEXES[case]]
    if isinstance(case, str):
        return case

    if plural:
        gender = case[GENDER_PLURAL_INDEXES['plural']]
    else:
        gender = case[GENDER_PLURAL_INDEXES[gender]]
    if isinstance(gender, str):
        return gender

    if animate:
        return gender[0]
    return gender[1]


# format:
# {n : [case_1 .. case_5]}
# case: text or [gender_1 .. gender_3 plural_4]
# gender: text or [animate, inanimate]
ONES = {
    0: ['ноль', 'ноля', 'нолю', 'ноль', 'нолём', 'ноле'],
    1: [['один', 'одна', 'одно', 'одни'],
        ['одного', 'одной', 'одного', 'одних'],
        ['одному', 'одной', 'одному', 'одним'],
        [['одного', 'один'], 'одну', 'одно', ['одних', 'одни']],
        ['одним', 'одной', 'одним', 'одними'],
        ['одном', 'одной', 'одном', 'одних']],
    2: [['два', 'две', 'два', 'двое'],
        ['двух'] * 3 + ['двоих'],
        ['двум'] * 3 + ['двоим'],
        [['двух', 'два'], ['двух', 'две'], 'два', 'двоих'],
        ['двумя'] * 3 + ['двоими'],
        ['двух'] * 3 + ['двоих']],
    3: [['три'] * 3 + ['трое'],
        ['трёх'] * 3 + ['троих'],
        ['трём'] * 3 + ['троим'],
        [['трёх', 'три'], ['трёх', 'три'], 'три', 'троих'],
        ['тремя'] * 3 + ['троими'],
        ['трёх'] * 3 + ['троих']],
    4: [['четыре'] * 3 + ['четверо'],
        ['четырёх'] * 3 + ['четверых'],
        ['четырём'] * 3 + ['четверым'],
        [['четырёх', 'четыре'], ['четырёх', 'четыре'], 'четыре', 'четверых'],
        ['четырьмя'] * 3 + ['четверыми'],
        ['четырёх'] * 3 + ['четверых']],
    5: ['пять', 'пяти', 'пяти', 'пять', 'пятью', 'пяти'],
    6: ['шесть', 'шести', 'шести', 'шесть', 'шестью', 'шести'],
    7: ['семь', 'семи', 'семи', 'семь', 'семью', 'семи'],
    8: ['восемь', 'восьми', 'восьми', 'восемь', 'восемью', 'восьми'],
    9: ['девять', 'девяти', 'девяти', 'девять', 'девятью', 'девяти']
}

ONES_ORD_PREFIXES = {0: 'нулев', 1: 'перв', 2: 'втор', 4: 'четвёрт', 5: 'пят',
                     6: 'шест', 7: 'седьм', 8: 'восьм', 9: 'девят'}
ONES_ORD_POSTFIXES_GROUPS = {0: 0, 1: 1, 2: 0, 4: 1, 5: 1, 6: 0, 7: 0, 8: 0,
                             9: 1}
CASE_POSTFIXES = [[{0: 'ой', 1: 'ый'}, 'ая', 'ое', 'ые'],
                  ['ого', 'ой', 'ого', 'ых'],
                  ['ому', 'ой', 'ому', 'ым'],
                  [['ого', {0: 'ой', 1: 'ый'}], 'ую', 'ое', ['ых', 'ые']],
                  ['ым', 'ой', 'ым', 'ыми'],
                  ['ом', 'ой', 'ом', 'ых']]


def get_cases(prefix, post_group):
    return [[
        prefix + postfix if isinstance(postfix, str) else
        [prefix + animate if isinstance(animate, str) else
         prefix + animate[post_group]
         for animate in postfix] if isinstance(postfix, list) else
        prefix + postfix[post_group]
        for postfix in case]
        for case in CASE_POSTFIXES]


def get_ord_classifier(prefixes, post_groups):
    if isinstance(post_groups, int):
        post_groups = {n: post_groups for n, i in prefixes.items()}
    return {
        num: get_cases(prefix, post_groups[num])
        for num, prefix in prefixes.items()
    }


ONES_ORD = {
    3: [['третий', 'третья', 'третье', 'третьи'],
        ['третьего', 'третьей', 'третьего', 'третьих'],
        ['третьему', 'третьей', 'третьему', 'третьим'],
        [['третьего', 'третий'], 'третью', 'третье', ['третьих', 'третьи']],
        ['третьим', 'третьей', 'третьим', 'третьими'],
        ['третьем', 'третьей', 'третьем', 'третьих']],
}
ONES_ORD.update(
    get_ord_classifier(ONES_ORD_PREFIXES, ONES_ORD_POSTFIXES_GROUPS)
)

TENS_PREFIXES = {1: 'один', 2: 'две', 3: 'три', 4: 'четыр', 5: 'пят',
                 6: 'шест', 7: 'сем', 8: 'восем', 9: 'девят'}
TENS_POSTFIXES = ['надцать', 'надцати', 'надцати', 'надцать', 'надцатью',
                  'надцати']
TENS = {0: ['десять', 'десяти', 'десяти', 'десять', 'десятью', 'десяти']}
TENS.update({
    num: [prefix + postfix for postfix in TENS_POSTFIXES]
    for num, prefix in TENS_PREFIXES.items()
})

TENS_ORD_PREFIXES = {0: "десят"}
TENS_ORD_PREFIXES.update({
    num: prefix + 'надцат' for num, prefix in TENS_PREFIXES.items()
})
TENS_ORD = get_ord_classifier(TENS_ORD_PREFIXES, 1)

TWENTIES = {
    2: ['двадцать', 'двадцати', 'двадцати', 'двадцать', 'двадцатью',
        'двадцати'],
    3: ['тридцать', 'тридцати', 'тридцати', 'тридцать', 'тридцатью',
        'тридцати'],
    4: ['сорок', 'сорока', 'сорока', 'сорок', 'сорока', 'сорока'],
    5: ['пятьдесят', 'пятидесяти', 'пятидесяти', 'пятьдесят', 'пятьюдесятью',
        'пятидесяти'],
    6: ['шестьдесят', 'шестидесяти', 'шестидесяти', 'шестьдесят',
        'шестьюдесятью', 'шестидесяти'],
    7: ['семьдесят', 'семидесяти', 'семидесяти', 'семьдесят', 'семьюдесятью',
        'семидесяти'],
    8: ['восемьдесят', 'восьмидесяти', 'восьмидесяти', 'восемьдесят',
        'восемьюдесятью', 'восьмидесяти'],
    9: ['девяносто', 'девяноста', 'девяноста', 'девяносто', 'девяноста',
        'девяноста'],
}

TWENTIES_ORD_PREFIXES = {2: 'двадцат', 3: 'тридцат', 4: 'сороков',
                         5: 'пятидесят', 6: 'шестидесят', 7: 'семидесят',
                         8: 'восьмидесят', 9: 'девяност'}
TWENTIES_ORD_POSTFIXES_GROUPS = {2: 1, 3: 1, 4: 0, 5: 1, 6: 1, 7: 1, 8: 1,
                                 9: 1}
TWENTIES_ORD = get_ord_classifier(TWENTIES_ORD_PREFIXES,
                                  TWENTIES_ORD_POSTFIXES_GROUPS)

HUNDREDS = {
    1: ['сто', 'ста', 'ста', 'сто', 'ста', 'ста'],
    2: ['двести', 'двухсот', 'двумстам', 'двести', 'двумястами', 'двухстах'],
    3: ['триста', 'трёхсот', 'трёмстам', 'триста', 'тремястами', 'трёхстах'],
    4: ['четыреста', 'четырёхсот', 'четырёмстам', 'четыреста', 'четырьмястами',
        'четырёхстах'],
    5: ['пятьсот', 'пятисот', 'пятистам', 'пятьсот', 'пятьюстами', 'пятистах'],
    6: ['шестьсот', 'шестисот', 'шестистам', 'шестьсот', 'шестьюстами',
        'шестистах'],
    7: ['семьсот', 'семисот', 'семистам', 'семьсот', 'семьюстами', 'семистах'],
    8: ['восемьсот', 'восьмисот', 'восьмистам', 'восемьсот', 'восемьюстами',
        'восьмистах'],
    9: ['девятьсот', 'девятисот', 'девятистам', 'девятьсот', 'девятьюстами',
        'девятистах'],
}

HUNDREDS_ORD_PREFIXES = {
    num: case[1] if num != 1 else 'сот' for num, case in HUNDREDS.items()
}
HUNDREDS_ORD = get_ord_classifier(HUNDREDS_ORD_PREFIXES, 1)


THOUSANDS_PREFIXES = {2: 'миллион', 3: 'миллиард', 4: 'триллион',
                      5: 'квадриллион', 6: 'квинтиллион', 7: 'секстиллион',
                      8: 'септиллион', 9: 'октиллион', 10: 'нониллион'}
THOUSANDS_POSTFIXES = [('', 'а', 'ов'),
                       ('а', 'ов', 'ов'),
                       ('у', 'ам', 'ам'),
                       ('', 'а', 'ов'),
                       ('ом', 'ами', 'ами'),
                       ('е', 'ах', 'ах')]
THOUSANDS = {
    1: [['тысяча', 'тысячи', 'тысяч'],
        ['тысячи', 'тысяч', 'тысяч'],
        ['тысяче', 'тысячам', 'тысячам'],
        ['тысячу', 'тысячи', 'тысяч'],
        ['тысячей', 'тысячами', 'тысячами'],
        ['тысяче', 'тысячах', 'тысячах']]
}
THOUSANDS.update({
    num: [
        [prefix + postfix for postfix in case] for case in THOUSANDS_POSTFIXES
    ] for num, prefix in THOUSANDS_PREFIXES.items()
})


def get_thousands_elements(num, case):
    return THOUSANDS[num][CASE_INDEXES[case]]


THOUSANDS_ORD_PREFIXES = {1: 'тысячн'}
THOUSANDS_ORD_PREFIXES.update({
    num: prefix + 'н' for num, prefix in THOUSANDS_PREFIXES.items()
})
THOUSANDS_ORD = get_ord_classifier(THOUSANDS_ORD_PREFIXES, 1)


class Num2Word_RU(Num2Word_Base):
    CURRENCY_FORMS = {
        'RUB': (
            ('рубль', 'рубля', 'рублей'), ('копейка', 'копейки', 'копеек')
        ),
        'EUR': (
            ('евро', 'евро', 'евро'), ('цент', 'цента', 'центов')
        ),
        'USD': (
            ('доллар', 'доллара', 'долларов'), ('цент', 'цента', 'центов')
        ),
        'UAH': (
            ('гривна', 'гривны', 'гривен'), ('копейка', 'копейки', 'копеек')
        ),
        'KZT': (
            ('тенге', 'тенге', 'тенге'), ('тиын', 'тиына', 'тиынов')
        ),
        'BYN': (
            ('белорусский рубль', 'белорусских рубля', 'белорусских рублей'),
            ('копейка', 'копейки', 'копеек')
        ),
        'UZS': (
            ('сум', 'сума', 'сумов'), ('тийин', 'тийина', 'тийинов')
        ),
        'PLN': (
            ('польский злотый', 'польских слотых', 'польских злотых'),
            ('грош', 'гроша', 'грошей')
        ),
    }

    def setup(self):
        self.negword = "минус"
        self.pointword = ('целая', 'целых', 'целых')
        self.pointword_ord = get_cases("цел", 1)

    def to_cardinal(self, number, case=D_CASE, plural=D_PLURAL,
                    gender=D_GENDER, animate=D_ANIMATE):
        n = str(number).replace(',', '.')
        if '.' in n:
            left, right = n.split('.')
            decimal_part = self._int2word(int(right), cardinal=True,
                                          gender='f')
            return u'%s %s %s %s' % (
                self._int2word(int(left), cardinal=True, gender='f'),
                self.pluralize(int(left), self.pointword),
                decimal_part,
                self.__decimal_bitness(right)
            )
        else:
            return self._int2word(int(n), cardinal=True, case=case,
                                  plural=plural, gender=gender,
                                  animate=animate)

    def __decimal_bitness(self, n):
        if n[-1] == "1" and n[-2:] != "11":
            return self._int2word(10 ** len(n), cardinal=False, gender='f')
        return self._int2word(10 ** len(n), cardinal=False, case='g',
                              plural=True)

    def pluralize(self, n, forms):
        if n % 100 in (11, 12, 13, 14):
            return forms[2]
        if n % 10 == 1:
            return forms[0]
        if n % 10 in (2, 3, 4):
            return forms[1]
        return forms[2]

    def to_ordinal(self, number, case=D_CASE, plural=D_PLURAL, gender=D_GENDER,
                   animate=D_ANIMATE):
        self.verify_ordinal(number)
        n = str(number).replace(',', '.')
        return self._int2word(int(n), cardinal=False, case=case, plural=plural,
                              gender=gender, animate=animate)

    def _money_verbose(self, number, currency):
        if currency == 'UAH':
            return self._int2word(number, gender='f')
        return self._int2word(number, gender='m')

    def _cents_verbose(self, number, currency):
        if currency in ('UAH', 'RUB', 'BYN'):
            return self._int2word(number, gender='f')
        return self._int2word(number, gender='m')

    def _int2word(self, n, feminine=False, cardinal=True, case=D_CASE,
                  plural=D_PLURAL, gender=D_GENDER, animate=D_ANIMATE):
        """
        n: number
        feminine: not used - for backward compatibility
        cardinal:True - cardinal
                False - ordinal
        case:   'n' - nominative
                'g' - genitive
                'd' - dative
                'a' - accusative
                'i' - instrumental
                'p' - prepositional
        plural: True - plural
                False - singular
        gender: 'f' - masculine
                'm' - feminine
                'n' - neuter
        animate: True - animate
                 False - inanimate
        """
        # For backward compatibility
        if feminine:
            gender = 'f'

        kwargs = {'case': case, 'plural': plural, 'gender': gender,
                  'animate': animate}

        if n < 0:
            return ' '.join([self.negword, self._int2word(abs(n),
                                                          cardinal=cardinal,
                                                          **kwargs)])

        if n == 0:
            return get_num_element(ONES, 0, **kwargs) if cardinal else \
                   get_num_element(ONES_ORD, 0, **kwargs)

        words = []
        chunks = list(splitbyx(str(n), 3))
        ord_join = chunks[-1] == 0  # join in one word if ending on 'тысячный'
        i = len(chunks)
        rightest_nonzero_chunk_i = i - 1 - max(
            [i for i, e in enumerate(chunks) if e != 0])
        for x in chunks:
            chunk_words = []
            i -= 1

            if x == 0:
                continue

            n1, n2, n3 = get_digits(x)

            if cardinal:
                chunk_words.extend(
                    self.__chunk_cardianl(n3, n2, n1, i, **kwargs)
                )
                if i > 0:
                    chunk_words.append(
                        self.pluralize(x, get_thousands_elements(i, case)))
            # ordinal, not joined like 'двухтысячный'
            elif not (ord_join and rightest_nonzero_chunk_i == i):
                chunk_words.extend(
                    self.__chunk_ordinal(n3, n2, n1, i, **kwargs)
                )
                if i > 0:
                    t_case = case if rightest_nonzero_chunk_i == i else 'n'
                    chunk_words.append(
                        self.pluralize(x, get_thousands_elements(i, t_case)))
            # ordinal, joined
            else:
                chunk_words.extend(
                    self.__chunk_ordinal_join(n3, n2, n1, i, **kwargs)
                )
                if i > 0:
                    chunk_words.append(
                        get_num_element(THOUSANDS_ORD, i, **kwargs))

                chunk_words = [''.join(chunk_words)]

            words.extend(chunk_words)

        return ' '.join(words)

    def __chunk_cardianl(self, hundreds, tens, ones, chunk_num, **kwargs):
        words = []
        if hundreds > 0:
            words.append(get_num_element(HUNDREDS, hundreds, **kwargs))

        if tens > 1:
            words.append(get_num_element(TWENTIES, tens, **kwargs))

        if tens == 1:
            words.append(get_num_element(TENS, ones, **kwargs))
        elif ones > 0:
            if chunk_num == 0:
                w_ones = get_num_element(ONES, ones, **kwargs)
            elif chunk_num == 1:
                # Thousands are feminine
                f_kwargs = kwargs.copy()
                f_kwargs['gender'] = 'f'
                w_ones = get_num_element(ONES, ones, **f_kwargs)
            else:
                w_ones = get_num_element(ONES, ones, **kwargs)

            words.append(w_ones)
        return words

    def __chunk_ordinal(self, hundreds, tens, ones, chunk_num, **kwargs):
        words = []
        if hundreds > 0:
            if tens == 0 and ones == 0:
                words.append(get_num_element(HUNDREDS_ORD, hundreds, **kwargs))
            else:
                words.append(get_num_element(HUNDREDS, hundreds))

        if tens > 1:
            if ones == 0:
                words.append(get_num_element(TWENTIES_ORD, tens, **kwargs))
            else:
                words.append(get_num_element(TWENTIES, tens))

        if tens == 1:
            words.append(get_num_element(TENS_ORD, ones, **kwargs))
        elif ones > 0:
            if chunk_num == 0:
                w_ones = get_num_element(ONES_ORD, ones, **kwargs)
            # тысячный, миллионнный и т.д.
            elif chunk_num > 0 and ones == 1 and hundreds == 0 and tens == 0:
                w_ones = None
            elif chunk_num == 1:
                # Thousands are feminine
                w_ones = get_num_element(ONES, ones, gender='f')
            else:
                w_ones = get_num_element(ONES, ones)

            if w_ones:
                words.append(w_ones)

        return words

    def __chunk_ordinal_join(self, hundreds, tens, ones, chunk_num, **kwargs):
        words = []

        if hundreds > 1:
            words.append(get_num_element(HUNDREDS, hundreds, case='g'))
        elif hundreds == 1:
            words.append(get_num_element(HUNDREDS, hundreds))   # стО, not стА

        if tens == 9:
            words.append(get_num_element(TWENTIES, tens))   # девяностО, not А
        elif tens > 1:
            words.append(get_num_element(TWENTIES, tens, case='g'))

        if tens == 1:
            words.append(get_num_element(TENS, ones, case='g'))
        elif ones > 0:
            if chunk_num == 0:
                w_ones = get_num_element(ONES_ORD, ones, **kwargs)
            # тысячный, миллионнный и т.д., двадцатиодномиллионный
            elif chunk_num > 0 and ones == 1 and tens != 1:
                if tens == 0 and hundreds == 0:
                    w_ones = None
                else:
                    w_ones = get_num_element(ONES, 1, gender='n')
            else:
                w_ones = get_num_element(ONES, ones, case='g')

            if w_ones:
                words.append(w_ones)

        return words



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base
from .utils import get_digits, splitbyx

ZERO = ('nula',)

ONES = {
    1: ('jedna',),
    2: ('dva',),
    3: ('tři',),
    4: ('čtyři',),
    5: ('pět',),
    6: ('šest',),
    7: ('sedm',),
    8: ('osm',),
    9: ('devět',),
}

TENS = {
    0: ('deset',),
    1: ('jedenáct',),
    2: ('dvanáct',),
    3: ('třináct',),
    4: ('čtrnáct',),
    5: ('patnáct',),
    6: ('šestnáct',),
    7: ('sedmnáct',),
    8: ('osmnáct',),
    9: ('devatenáct',),
}

TWENTIES = {
    2: ('dvacet',),
    3: ('třicet',),
    4: ('čtyřicet',),
    5: ('padesát',),
    6: ('šedesát',),
    7: ('sedmdesát',),
    8: ('osmdesát',),
    9: ('devadesát',),
}

HUNDREDS = {
    1: ('sto',),
    2: ('dvěstě',),
    3: ('třista',),
    4: ('čtyřista',),
    5: ('pětset',),
    6: ('šestset',),
    7: ('sedmset',),
    8: ('osmset',),
    9: ('devětset',),
}

THOUSANDS = {
    1: ('tisíc', 'tisíce', 'tisíc'),  # 10^3
    2: ('milion', 'miliony', 'milionů'),  # 10^6
    3: ('miliarda', 'miliardy', 'miliard'),  # 10^9
    4: ('bilion', 'biliony', 'bilionů'),  # 10^12
    5: ('biliarda', 'biliardy', 'biliard'),  # 10^15
    6: ('trilion', 'triliony', 'trilionů'),  # 10^18
    7: ('triliarda', 'triliardy', 'triliard'),  # 10^21
    8: ('kvadrilion', 'kvadriliony', 'kvadrilionů'),  # 10^24
    9: ('kvadriliarda', 'kvadriliardy', 'kvadriliard'),  # 10^27
    10: ('quintillion', 'quintilliony', 'quintillionů'),  # 10^30
}


class Num2Word_CS(Num2Word_Base):
    CURRENCY_FORMS = {
        'CZK': (
            ('koruna', 'koruny', 'korun'), ('halíř', 'halíře', 'haléřů')
        ),
        'EUR': (
            ('euro', 'euro', 'euro'), ('cent', 'centy', 'centů')
        ),
    }

    def setup(self):
        self.negword = "mínus"
        self.pointword = "celá"

    def to_cardinal(self, number):
        n = str(number).replace(',', '.')
        if '.' in n:
            left, right = n.split('.')
            leading_zero_count = len(right) - len(right.lstrip('0'))
            decimal_part = ((ZERO[0] + ' ') * leading_zero_count +
                            self._int2word(int(right)))
            return u'%s %s %s' % (
                self._int2word(int(left)),
                self.pointword,
                decimal_part
            )
        else:
            return self._int2word(int(n))

    def pluralize(self, n, forms):
        if n == 1:
            form = 0
        elif 5 > n % 10 > 1 and (n % 100 < 10 or n % 100 > 20):
            form = 1
        else:
            form = 2
        return forms[form]

    def to_ordinal(self, number):
        raise NotImplementedError()

    def _int2word(self, n):
        if n == 0:
            return ZERO[0]

        words = []
        chunks = list(splitbyx(str(n), 3))
        i = len(chunks)
        for x in chunks:
            i -= 1

            if x == 0:
                continue

            n1, n2, n3 = get_digits(x)

            if n3 > 0:
                words.append(HUNDREDS[n3][0])

            if n2 > 1:
                words.append(TWENTIES[n2][0])

            if n2 == 1:
                words.append(TENS[n1][0])
            elif n1 > 0 and not (i > 0 and x == 1):
                words.append(ONES[n1][0])

            if i > 0:
                words.append(self.pluralize(x, THOUSANDS[i]))

        return ' '.join(words)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU


class Num2Word_SV(lang_EU.Num2Word_EU):
    GIGA_SUFFIX = "iljarder"
    MEGA_SUFFIX = "iljoner"

    def set_high_numwords(self, high):
        cap = 3 + 6 * len(high)

        for word, n in zip(high, range(cap, 3, -6)):
            if self.GIGA_SUFFIX:
                self.cards[10 ** n] = word + self.GIGA_SUFFIX

            if self.MEGA_SUFFIX:
                self.cards[10 ** (n - 3)] = word + self.MEGA_SUFFIX

    def setup(self):
        super(Num2Word_SV, self).setup()

        self.negword = "minus "
        self.pointword = "komma"
        self.exclude_title = ["och", "komma", "minus"]

        self.mid_numwords = [(1000, "tusen"), (100, "hundra"),
                             (90, "nittio"), (80, "åttio"), (70, "sjuttio"),
                             (60, "sextio"), (50, "femtio"), (40, "förtio"),
                             (30, "trettio")]
        self.low_numwords = ["tjugo", "nitton", "arton", "sjutton",
                             "sexton", "femton", "fjorton", "tretton",
                             "tolv", "elva", "tio", "nio", "åtta",
                             "sju", "sex", "fem", "fyra", "tre", "två",
                             "ett", "noll"]
        self.ords = {"noll": "nollte",
                     "ett": "första",
                     "två": "andra",
                     "tre": "tredje",
                     "fyra": "fjärde",
                     "fem": "femte",
                     "sex": "sjätte",
                     "sju": "sjunde",
                     "åtta": "åttonde",
                     "nio": "nionde",
                     "tio": "tionde",
                     "elva": "elfte",
                     "tolv": "tolfte",
                     "tjugo": "tjugonde"}

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif 100 > lnum > rnum:
            return ("%s%s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 100 > rnum:
            return ("%s%s" % (ltext, rtext), lnum + rnum)
        elif rnum >= 1000000 and lnum == 1:
            return ("%s %s" % ('en', rtext[:-2]), lnum + rnum)
        elif rnum >= 1000000 and lnum > 1:
            return ("%s %s" % (ltext, rtext), lnum + rnum)
        elif rnum > lnum:
            return ("%s%s" % (ltext, rtext), lnum * rnum)
        return ("%s %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outwords = self.to_cardinal(value).split(" ")
        lastword = outwords[-1]
        ending_length = 0
        try:
            lastword_ending = self.ords[lastword[-4:]]
            ending_length = 4
        except KeyError:
            try:
                lastword_ending = self.ords[lastword[-3:]]
                ending_length = 3
            except KeyError:
                lastword_ending = "de"
        if lastword_ending == 'de':
            lastword_first_part = self.title(lastword)[:]
        else:
            lastword_first_part = self.title(lastword)[:-ending_length]
        lastword_correct = lastword_first_part + lastword_ending
        outwords[-1] = lastword_correct
        return " ".join(outwords)

    def to_ordinal_num(self, value):
        raise NotImplementedError(
            "'ordinal_num' is not implemented for swedish language")

    def to_year(self, val, longval=True):
        raise NotImplementedError(
            "'year' is not implemented for swedish language")

    def to_currency(self, val, longval=True):
        raise NotImplementedError(
            "'currency' is not implemented for swedish language")



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import print_function, unicode_literals

from .lang_ES import Num2Word_ES


class Num2Word_ES_NI(Num2Word_ES):
    CURRENCY_FORMS = {
        'NIO': (('córdoba', 'córdobas'), ('centavo', 'centavos')),
    }

    def to_currency(self, val, currency='NIO', cents=True, separator=' con',
                    adjective=False):
        result = super(Num2Word_ES, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)
        return result.replace("uno", "un")
